@@ -5,23 +5,20 @@ import (
 	"fmt"
 	"net"
 	"strconv"
-	"strings"
 	"sync"
 	"sync/atomic"
 	"time"
 
 	"github.com/armon/go-metrics"
 	"github.com/hashicorp/consul/acl"
-	"github.com/hashicorp/consul/agent/connect"
-	ca "github.com/hashicorp/consul/agent/connect/ca"
 	"github.com/hashicorp/consul/agent/consul/autopilot"
 	"github.com/hashicorp/consul/agent/metadata"
 	"github.com/hashicorp/consul/agent/structs"
 	"github.com/hashicorp/consul/api"
 	"github.com/hashicorp/consul/lib"
 	"github.com/hashicorp/consul/types"
-	memdb "github.com/hashicorp/go-memdb"
-	uuid "github.com/hashicorp/go-uuid"
+	"github.com/hashicorp/go-memdb"
+	"github.com/hashicorp/go-uuid"
 	"github.com/hashicorp/go-version"
 	"github.com/hashicorp/raft"
 	"github.com/hashicorp/serf/serf"
@@ -40,6 +37,10 @@ var (
 	// minAutopilotVersion is the minimum Consul version in which Autopilot features
 	// are supported.
 	minAutopilotVersion = version.Must(version.NewVersion("0.8.0"))
+
+	// minCentralizedConfigVersion is the minimum Consul version in which centralized
+	// config is supported
+	minCentralizedConfigVersion = version.Must(version.NewVersion("1.5.0"))
 )
 
 // monitorLeadership is used to monitor if we acquire or lose our role
@@ -122,6 +123,21 @@ func (s *Server) monitorLeadership() {
 	}
 }
 
+func (s *Server) leadershipTransfer() error {
+	retryCount := 3
+	for i := 0; i < retryCount; i++ {
+		future := s.raft.LeadershipTransfer()
+		if err := future.Error(); err != nil {
+			s.logger.Printf("[ERR] consul: failed to transfer leadership attempt %d/%d: %v", i, retryCount, err)
+		} else {
+			s.logger.Printf("[ERR] consul: successfully transferred leadership attempt %d/%d", i, retryCount)
+			return nil
+		}
+
+	}
+	return fmt.Errorf("failed to transfer leadership in %d attempts", retryCount)
+}
+
 // leaderLoop runs as long as we are the leader to run various
 // maintenance activities
 func (s *Server) leaderLoop(stopCh chan struct{}) {
@@ -138,19 +154,6 @@ func (s *Server) leaderLoop(stopCh chan struct{}) {
 	var reconcileCh chan serf.Member
 	establishedLeader := false
 
-	reassert := func() error {
-		if !establishedLeader {
-			return fmt.Errorf("leadership has not been established")
-		}
-		if err := s.revokeLeadership(); err != nil {
-			return err
-		}
-		if err := s.establishLeadership(); err != nil {
-			return err
-		}
-		return nil
-	}
-
 RECONCILE:
 	// Setup a reconciliation timer
 	reconcileCh = nil
@@ -171,17 +174,22 @@ RECONCILE:
 			s.logger.Printf("[ERR] consul: failed to establish leadership: %v", err)
 			// Immediately revoke leadership since we didn't successfully
 			// establish leadership.
-			if err := s.revokeLeadership(); err != nil {
-				s.logger.Printf("[ERR] consul: failed to revoke leadership: %v", err)
+			s.revokeLeadership()
+
+			// attempt to transfer leadership. If successful it is
+			// time to leave the leaderLoop since this node is no
+			// longer the leader. If leadershipTransfer() fails, we
+			// will try to acquire it again after
+			// 5 seconds.
+			if err := s.leadershipTransfer(); err != nil {
+				s.logger.Printf("[ERR] consul: %v", err)
+				interval = time.After(5 * time.Second)
+				goto WAIT
 			}
-			goto WAIT
+			return
 		}
 		establishedLeader = true
-		defer func() {
-			if err := s.revokeLeadership(); err != nil {
-				s.logger.Printf("[ERR] consul: failed to revoke leadership: %v", err)
-			}
-		}()
+		defer s.revokeLeadership()
 	}
 
 	// Reconcile any missing data
@@ -219,7 +227,47 @@ WAIT:
 		case index := <-s.tombstoneGC.ExpireCh():
 			go s.reapTombstones(index)
 		case errCh := <-s.reassertLeaderCh:
-			errCh <- reassert()
+			// we can get into this state when the initial
+			// establishLeadership has failed as well as the follow
+			// up leadershipTransfer. Afterwards we will be waiting
+			// for the interval to trigger a reconciliation and can
+			// potentially end up here. There is no point to
+			// reassert because this agent was never leader in the
+			// first place.
+			if !establishedLeader {
+				errCh <- fmt.Errorf("leadership has not been established")
+				continue
+			}
+
+			// continue to reassert only if we previously were the
+			// leader, which means revokeLeadership followed by an
+			// establishLeadership().
+			s.revokeLeadership()
+			err := s.establishLeadership()
+			errCh <- err
+
+			// in case establishLeadership failed, we will try to
+			// transfer leadership. At this time raft thinks we are
+			// the leader, but consul disagrees.
+			if err != nil {
+				if err := s.leadershipTransfer(); err != nil {
+					// establishedLeader was true before,
+					// but it no longer is since it revoked
+					// leadership and Leadership transfer
+					// also failed. Which is why it stays
+					// in the leaderLoop, but now
+					// establishedLeader needs to be set to
+					// false.
+					establishedLeader = false
+					interval = time.After(5 * time.Second)
+					goto WAIT
+				}
+
+				// leadershipTransfer was successful and it is
+				// time to leave the leaderLoop.
+				return
+			}
+
 		}
 	}
 }
@@ -261,6 +309,11 @@ func (s *Server) establishLeadership() error {
 		return err
 	}
 
+	// attempt to bootstrap config entries
+	if err := s.bootstrapConfigEntries(s.config.ConfigEntryBootstrap); err != nil {
+		return err
+	}
+
 	s.getOrCreateAutopilotConfig()
 	s.autopilot.Start()
 
@@ -269,9 +322,9 @@ func (s *Server) establishLeadership() error {
 		return err
 	}
 
-	s.startEnterpriseLeader()
+	s.startConfigReplication()
 
-	s.startCARootPruning()
+	s.startConnectLeader()
 
 	s.setConsistentReadReady()
 	return nil
@@ -279,27 +332,27 @@ func (s *Server) establishLeadership() error {
 
 // revokeLeadership is invoked once we step down as leader.
 // This is used to cleanup any state that may be specific to a leader.
-func (s *Server) revokeLeadership() error {
+func (s *Server) revokeLeadership() {
 	// Disable the tombstone GC, since it is only useful as a leader
 	s.tombstoneGC.SetEnabled(false)
 
 	// Clear the session timers on either shutdown or step down, since we
 	// are no longer responsible for session expirations.
-	if err := s.clearAllSessionTimers(); err != nil {
-		return err
-	}
+	s.clearAllSessionTimers()
 
-	s.stopEnterpriseLeader()
+	s.stopConfigReplication()
 
-	s.stopCARootPruning()
+	s.stopConnectLeader()
 
 	s.setCAProvider(nil, nil)
 
+	s.stopACLTokenReaping()
+
 	s.stopACLUpgrade()
 
 	s.resetConsistentReadReady()
+
 	s.autopilot.Stop()
-	return nil
 }
 
 // DEPRECATED (ACL-Legacy-Compat) - Remove once old ACL compatibility is removed
@@ -316,6 +369,7 @@ func (s *Server) initializeLegacyACL() error {
 	if err != nil {
 		return fmt.Errorf("failed to get anonymous token: %v", err)
 	}
+	// Ignoring expiration times to avoid an insertion collision.
 	if token == nil {
 		req := structs.ACLRequest{
 			Datacenter: authDC,
@@ -339,6 +393,7 @@ func (s *Server) initializeLegacyACL() error {
 		if err != nil {
 			return fmt.Errorf("failed to get master token: %v", err)
 		}
+		// Ignoring expiration times to avoid an insertion collision.
 		if token == nil {
 			req := structs.ACLRequest{
 				Datacenter: authDC,
@@ -410,6 +465,25 @@ func (s *Server) initializeACLs(upgrade bool) error {
 	// leader.
 	s.acls.cache.Purge()
 
+	// Purge the auth method validators since they could've changed while we
+	// were not leader.
+	s.purgeAuthMethodValidators()
+
+	// Remove any token affected by CVE-2019-8336
+	if !s.InACLDatacenter() {
+		_, token, err := s.fsm.State().ACLTokenGetBySecret(nil, redactedToken)
+		if err == nil && token != nil {
+			req := structs.ACLTokenBatchDeleteRequest{
+				TokenIDs: []string{token.AccessorID},
+			}
+
+			_, err := s.raftApply(structs.ACLTokenDeleteRequestType, &req)
+			if err != nil {
+				return fmt.Errorf("failed to remove token with a redacted secret: %v", err)
+			}
+		}
+	}
+
 	if s.InACLDatacenter() {
 		if s.UseLegacyACLs() && !upgrade {
 			s.logger.Printf("[INFO] acl: initializing legacy acls")
@@ -454,6 +528,7 @@ func (s *Server) initializeACLs(upgrade bool) error {
 			if err != nil {
 				return fmt.Errorf("failed to get master token: %v", err)
 			}
+			// Ignoring expiration times to avoid an insertion collision.
 			if token == nil {
 				accessor, err := lib.GenerateUUID(s.checkTokenUUID)
 				if err != nil {
@@ -515,6 +590,7 @@ func (s *Server) initializeACLs(upgrade bool) error {
 		if err != nil {
 			return fmt.Errorf("failed to get anonymous token: %v", err)
 		}
+		// Ignoring expiration times to avoid an insertion collision.
 		if token == nil {
 			// DEPRECATED (ACL-Legacy-Compat) - Don't need to query for previous "anonymous" token
 			// check for legacy token that needs an upgrade
@@ -522,6 +598,7 @@ func (s *Server) initializeACLs(upgrade bool) error {
 			if err != nil {
 				return fmt.Errorf("failed to get anonymous token: %v", err)
 			}
+			// Ignoring expiration times to avoid an insertion collision.
 
 			// the token upgrade routine will take care of upgrading the token if a legacy version exists
 			if legacyToken == nil {
@@ -544,6 +621,7 @@ func (s *Server) initializeACLs(upgrade bool) error {
 				s.logger.Printf("[INFO] consul: Created ACL anonymous token from configuration")
 			}
 		}
+		// launch the upgrade go routine to generate accessors for everything
 		s.startACLUpgrade()
 	} else {
 		if s.UseLegacyACLs() && !upgrade {
@@ -560,7 +638,7 @@ func (s *Server) initializeACLs(upgrade bool) error {
 		s.startACLReplication()
 	}
 
-	// launch the upgrade go routine to generate accessors for everything
+	s.startACLTokenReaping()
 
 	return nil
 }
@@ -589,6 +667,7 @@ func (s *Server) startACLUpgrade() {
 			if err != nil {
 				s.logger.Printf("[WARN] acl: encountered an error while searching for tokens without accessor ids: %v", err)
 			}
+			// No need to check expiration time here, as that only exists for v2 tokens.
 
 			if len(tokens) == 0 {
 				ws := memdb.NewWatchSet()
@@ -621,7 +700,10 @@ func (s *Server) startACLUpgrade() {
 				}
 
 				// Assign the global-management policy to legacy management tokens
-				if len(newToken.Policies) == 0 && newToken.Type == structs.ACLTokenTypeManagement {
+				if len(newToken.Policies) == 0 &&
+					len(newToken.ServiceIdentities) == 0 &&
+					len(newToken.Roles) == 0 &&
+					newToken.Type == structs.ACLTokenTypeManagement {
 					newToken.Policies = append(newToken.Policies, structs.ACLTokenPolicyLink{ID: structs.ACLPolicyGlobalManagementID})
 				}
 
@@ -684,7 +766,7 @@ func (s *Server) startLegacyACLReplication() {
 				return
 			}
 
-			if s.tokens.ACLReplicationToken() == "" {
+			if s.tokens.ReplicationToken() == "" {
 				continue
 			}
 
@@ -699,7 +781,7 @@ func (s *Server) startLegacyACLReplication() {
 				s.logger.Printf("[WARN] consul: Legacy ACL replication error (will retry if still leader): %v", err)
 			} else {
 				lastRemoteIndex = index
-				s.updateACLReplicationStatusIndex(index)
+				s.updateACLReplicationStatusIndex(structs.ACLReplicateLegacy, index)
 				s.logger.Printf("[DEBUG] consul: Legacy ACL replication completed through remote index %d", index)
 			}
 		}
@@ -721,8 +803,22 @@ func (s *Server) startACLReplication() {
 	ctx, cancel := context.WithCancel(context.Background())
 	s.aclReplicationCancel = cancel
 
-	replicationType := structs.ACLReplicatePolicies
+	s.startACLReplicator(ctx, structs.ACLReplicatePolicies, s.replicateACLPolicies)
+	s.startACLReplicator(ctx, structs.ACLReplicateRoles, s.replicateACLRoles)
 
+	if s.config.ACLTokenReplication {
+		s.startACLReplicator(ctx, structs.ACLReplicateTokens, s.replicateACLTokens)
+		s.updateACLReplicationStatusRunning(structs.ACLReplicateTokens)
+	} else {
+		s.updateACLReplicationStatusRunning(structs.ACLReplicatePolicies)
+	}
+
+	s.aclReplicationEnabled = true
+}
+
+type replicateFunc func(ctx context.Context, lastRemoteIndex uint64) (uint64, bool, error)
+
+func (s *Server) startACLReplicator(ctx context.Context, replicationType structs.ACLReplicationType, replicateFunc replicateFunc) {
 	go func() {
 		var failedAttempts uint
 		limiter := rate.NewLimiter(rate.Limit(s.config.ACLReplicationRate), s.config.ACLReplicationBurst)
@@ -733,11 +829,11 @@ func (s *Server) startACLReplication() {
 				return
 			}
 
-			if s.tokens.ACLReplicationToken() == "" {
+			if s.tokens.ReplicationToken() == "" {
 				continue
 			}
 
-			index, exit, err := s.replicateACLPolicies(lastRemoteIndex, ctx)
+			index, exit, err := replicateFunc(ctx, lastRemoteIndex)
 			if exit {
 				return
 			}
@@ -745,7 +841,7 @@ func (s *Server) startACLReplication() {
 			if err != nil {
 				lastRemoteIndex = 0
 				s.updateACLReplicationStatusError()
-				s.logger.Printf("[WARN] consul: ACL policy replication error (will retry if still leader): %v", err)
+				s.logger.Printf("[WARN] consul: ACL %s replication error (will retry if still leader): %v", replicationType.SingularNoun(), err)
 				if (1 << failedAttempts) < aclReplicationMaxRetryBackoff {
 					failedAttempts++
 				}
@@ -758,65 +854,14 @@ func (s *Server) startACLReplication() {
 				}
 			} else {
 				lastRemoteIndex = index
-				s.updateACLReplicationStatusIndex(index)
-				s.logger.Printf("[DEBUG] consul: ACL policy replication completed through remote index %d", index)
+				s.updateACLReplicationStatusIndex(replicationType, index)
+				s.logger.Printf("[DEBUG] consul: ACL %s replication completed through remote index %d", replicationType.SingularNoun(), index)
 				failedAttempts = 0
 			}
 		}
 	}()
 
-	s.logger.Printf("[INFO] acl: started ACL Policy replication")
-
-	if s.config.ACLTokenReplication {
-		replicationType = structs.ACLReplicateTokens
-
-		go func() {
-			var failedAttempts uint
-			limiter := rate.NewLimiter(rate.Limit(s.config.ACLReplicationRate), s.config.ACLReplicationBurst)
-			var lastRemoteIndex uint64
-			for {
-				if err := limiter.Wait(ctx); err != nil {
-					return
-				}
-
-				if s.tokens.ACLReplicationToken() == "" {
-					continue
-				}
-
-				index, exit, err := s.replicateACLTokens(lastRemoteIndex, ctx)
-				if exit {
-					return
-				}
-
-				if err != nil {
-					lastRemoteIndex = 0
-					s.updateACLReplicationStatusError()
-					s.logger.Printf("[WARN] consul: ACL token replication error (will retry if still leader): %v", err)
-					if (1 << failedAttempts) < aclReplicationMaxRetryBackoff {
-						failedAttempts++
-					}
-
-					select {
-					case <-ctx.Done():
-						return
-					case <-time.After((1 << failedAttempts) * time.Second):
-						// do nothing
-					}
-				} else {
-					lastRemoteIndex = index
-					s.updateACLReplicationStatusTokenIndex(index)
-					s.logger.Printf("[DEBUG] consul: ACL token replication completed through remote index %d", index)
-					failedAttempts = 0
-				}
-			}
-		}()
-
-		s.logger.Printf("[INFO] acl: started ACL Token replication")
-	}
-
-	s.updateACLReplicationStatusRunning(replicationType)
-
-	s.aclReplicationEnabled = true
+	s.logger.Printf("[INFO] acl: started ACL %s replication", replicationType.SingularNoun())
 }
 
 func (s *Server) stopACLReplication() {
@@ -833,6 +878,20 @@ func (s *Server) stopACLReplication() {
 	s.aclReplicationEnabled = false
 }
 
+func (s *Server) startConfigReplication() {
+	if s.config.PrimaryDatacenter == "" || s.config.PrimaryDatacenter == s.config.Datacenter {
+		// replication shouldn't run in the primary DC
+		return
+	}
+
+	s.configReplicator.Start()
+}
+
+func (s *Server) stopConfigReplication() {
+	// will be a no-op when not started
+	s.configReplicator.Stop()
+}
+
 // getOrCreateAutopilotConfig is used to get the autopilot config, initializing it if necessary
 func (s *Server) getOrCreateAutopilotConfig() *autopilot.Config {
 	state := s.fsm.State()
@@ -860,272 +919,54 @@ func (s *Server) getOrCreateAutopilotConfig() *autopilot.Config {
 	return config
 }
 
-// initializeCAConfig is used to initialize the CA config if necessary
-// when setting up the CA during establishLeadership
-func (s *Server) initializeCAConfig() (*structs.CAConfiguration, error) {
-	state := s.fsm.State()
-	_, config, err := state.CAConfig()
-	if err != nil {
-		return nil, err
-	}
-	if config != nil {
-		return config, nil
-	}
-
-	config = s.config.CAConfig
-	if config.ClusterID == "" {
-		id, err := uuid.GenerateUUID()
-		if err != nil {
-			return nil, err
-		}
-		config.ClusterID = id
-	}
-
-	req := structs.CARequest{
-		Op:     structs.CAOpSetConfig,
-		Config: config,
-	}
-	if _, err = s.raftApply(structs.ConnectCARequestType, req); err != nil {
-		return nil, err
-	}
-
-	return config, nil
-}
-
-// initializeRootCA runs the initialization logic for a root CA.
-func (s *Server) initializeRootCA(provider ca.Provider, conf *structs.CAConfiguration) error {
-	if err := provider.Configure(conf.ClusterID, true, conf.Config); err != nil {
-		return fmt.Errorf("error configuring provider: %v", err)
-	}
-	if err := provider.GenerateRoot(); err != nil {
-		return fmt.Errorf("error generating CA root certificate: %v", err)
-	}
-
-	// Get the active root cert from the CA
-	rootPEM, err := provider.ActiveRoot()
-	if err != nil {
-		return fmt.Errorf("error getting root cert: %v", err)
-	}
-
-	rootCA, err := parseCARoot(rootPEM, conf.Provider, conf.ClusterID)
-	if err != nil {
-		return err
-	}
-
-	// Check if the CA root is already initialized and exit if it is,
-	// adding on any existing intermediate certs since they aren't directly
-	// tied to the provider.
-	// Every change to the CA after this initial bootstrapping should
-	// be done through the rotation process.
-	state := s.fsm.State()
-	_, activeRoot, err := state.CARootActive(nil)
-	if err != nil {
-		return err
-	}
-	if activeRoot != nil {
-		// This state shouldn't be possible to get into because we update the root and
-		// CA config in the same FSM operation.
-		if activeRoot.ID != rootCA.ID {
-			return fmt.Errorf("stored CA root %q is not the active root (%s)", rootCA.ID, activeRoot.ID)
-		}
-
-		rootCA.IntermediateCerts = activeRoot.IntermediateCerts
-		s.setCAProvider(provider, rootCA)
-
+func (s *Server) bootstrapConfigEntries(entries []structs.ConfigEntry) error {
+	if s.config.PrimaryDatacenter != "" && s.config.PrimaryDatacenter != s.config.Datacenter {
+		// only bootstrap in the primary datacenter
 		return nil
 	}
 
-	// Get the highest index
-	idx, _, err := state.CARoots(nil)
-	if err != nil {
-		return err
+	if len(entries) < 1 {
+		// nothing to initialize
+		return nil
 	}
 
-	// Store the root cert in raft
-	resp, err := s.raftApply(structs.ConnectCARequestType, &structs.CARequest{
-		Op:    structs.CAOpSetRoots,
-		Index: idx,
-		Roots: []*structs.CARoot{rootCA},
-	})
-	if err != nil {
-		s.logger.Printf("[ERR] connect: Apply failed %v", err)
-		return err
-	}
-	if respErr, ok := resp.(error); ok {
-		return respErr
+	if !ServersMeetMinimumVersion(s.LANMembers(), minCentralizedConfigVersion) {
+		s.logger.Printf("[WARN] centralized config: can't initialize until all servers >= %s", minCentralizedConfigVersion.String())
+		return nil
 	}
 
-	s.setCAProvider(provider, rootCA)
-
-	s.logger.Printf("[INFO] connect: initialized primary datacenter CA with provider %q", conf.Provider)
-
-	return nil
-}
-
-// parseCARoot returns a filled-in structs.CARoot from a raw PEM value.
-func parseCARoot(pemValue, provider, clusterID string) (*structs.CARoot, error) {
-	id, err := connect.CalculateCertFingerprint(pemValue)
-	if err != nil {
-		return nil, fmt.Errorf("error parsing root fingerprint: %v", err)
-	}
-	rootCert, err := connect.ParseCert(pemValue)
-	if err != nil {
-		return nil, fmt.Errorf("error parsing root cert: %v", err)
-	}
-	return &structs.CARoot{
-		ID:                  id,
-		Name:                fmt.Sprintf("%s CA Root Cert", strings.Title(provider)),
-		SerialNumber:        rootCert.SerialNumber.Uint64(),
-		SigningKeyID:        connect.HexString(rootCert.AuthorityKeyId),
-		ExternalTrustDomain: clusterID,
-		NotBefore:           rootCert.NotBefore,
-		NotAfter:            rootCert.NotAfter,
-		RootCert:            pemValue,
-		Active:              true,
-	}, nil
-}
-
-// createProvider returns a connect CA provider from the given config.
-func (s *Server) createCAProvider(conf *structs.CAConfiguration) (ca.Provider, error) {
-	switch conf.Provider {
-	case structs.ConsulCAProvider:
-		return &ca.ConsulProvider{Delegate: &consulCADelegate{s}}, nil
-	case structs.VaultCAProvider:
-		return &ca.VaultProvider{}, nil
-	default:
-		return nil, fmt.Errorf("unknown CA provider %q", conf.Provider)
-	}
-}
-
-func (s *Server) getCAProvider() (ca.Provider, *structs.CARoot) {
-	retries := 0
-	var result ca.Provider
-	var resultRoot *structs.CARoot
-	for result == nil {
-		s.caProviderLock.RLock()
-		result = s.caProvider
-		resultRoot = s.caProviderRoot
-		s.caProviderLock.RUnlock()
-
-		// In cases where an agent is started with managed proxies, we may ask
-		// for the provider before establishLeadership completes. If we're the
-		// leader, then wait and get the provider again
-		if result == nil && s.IsLeader() && retries < 10 {
-			retries++
-			time.Sleep(50 * time.Millisecond)
-			continue
+	state := s.fsm.State()
+	for _, entry := range entries {
+		// avoid a round trip through Raft if we know the CAS is going to fail
+		_, existing, err := state.ConfigEntry(nil, entry.GetKind(), entry.GetName())
+		if err != nil {
+			return fmt.Errorf("Failed to determine whether the configuration for %q / %q already exists: %v", entry.GetKind(), entry.GetName(), err)
 		}
 
-		break
-	}
+		if existing == nil {
+			// ensure the ModifyIndex is set to 0 for the CAS request
+			entry.GetRaftIndex().ModifyIndex = 0
 
-	return result, resultRoot
-}
+			req := structs.ConfigEntryRequest{
+				Op:         structs.ConfigEntryUpsertCAS,
+				Datacenter: s.config.Datacenter,
+				Entry:      entry,
+			}
 
-func (s *Server) setCAProvider(newProvider ca.Provider, root *structs.CARoot) {
-	s.caProviderLock.Lock()
-	defer s.caProviderLock.Unlock()
-	s.caProvider = newProvider
-	s.caProviderRoot = root
-}
-
-// startCARootPruning starts a goroutine that looks for stale CARoots
-// and removes them from the state store.
-func (s *Server) startCARootPruning() {
-	s.caPruningLock.Lock()
-	defer s.caPruningLock.Unlock()
-
-	if s.caPruningEnabled {
-		return
-	}
-
-	s.caPruningCh = make(chan struct{})
-
-	go func() {
-		ticker := time.NewTicker(caRootPruneInterval)
-		defer ticker.Stop()
-
-		for {
-			select {
-			case <-s.caPruningCh:
-				return
-			case <-ticker.C:
-				if err := s.pruneCARoots(); err != nil {
-					s.logger.Printf("[ERR] connect: error pruning CA roots: %v", err)
+			resp, err := s.raftApply(structs.ConfigEntryRequestType, &req)
+			if err == nil {
+				if respErr, ok := resp.(error); ok {
+					err = respErr
 				}
 			}
+			if err != nil {
+				return fmt.Errorf("Failed to apply configuration entry %q / %q: %v", entry.GetKind(), entry.GetName(), err)
+			}
 		}
-	}()
-
-	s.caPruningEnabled = true
-}
-
-// pruneCARoots looks for any CARoots that have been rotated out and expired.
-func (s *Server) pruneCARoots() error {
-	if !s.config.ConnectEnabled {
-		return nil
 	}
-
-	state := s.fsm.State()
-	idx, roots, err := state.CARoots(nil)
-	if err != nil {
-		return err
-	}
-
-	_, caConf, err := state.CAConfig()
-	if err != nil {
-		return err
-	}
-
-	common, err := caConf.GetCommonConfig()
-	if err != nil {
-		return err
-	}
-
-	var newRoots structs.CARoots
-	for _, r := range roots {
-		if !r.Active && !r.RotatedOutAt.IsZero() && time.Now().Sub(r.RotatedOutAt) > common.LeafCertTTL*2 {
-			s.logger.Printf("[INFO] connect: pruning old unused root CA (ID: %s)", r.ID)
-			continue
-		}
-		newRoot := *r
-		newRoots = append(newRoots, &newRoot)
-	}
-
-	// Return early if there's nothing to remove.
-	if len(newRoots) == len(roots) {
-		return nil
-	}
-
-	// Commit the new root state.
-	var args structs.CARequest
-	args.Op = structs.CAOpSetRoots
-	args.Index = idx
-	args.Roots = newRoots
-	resp, err := s.raftApply(structs.ConnectCARequestType, args)
-	if err != nil {
-		return err
-	}
-	if respErr, ok := resp.(error); ok {
-		return respErr
-	}
-
 	return nil
 }
 
-// stopCARootPruning stops the CARoot pruning process.
-func (s *Server) stopCARootPruning() {
-	s.caPruningLock.Lock()
-	defer s.caPruningLock.Unlock()
-
-	if !s.caPruningEnabled {
-		return
-	}
-
-	close(s.caPruningCh)
-	s.caPruningEnabled = false
-}
-
 // reconcileReaped is used to reconcile nodes that have failed and been reaped
 // from Serf but remain in the catalog. This is done by looking for unknown nodes with serfHealth checks registered.
 // We generate a "reap" event to cause the node to be cleaned up.
@@ -1255,6 +1096,17 @@ func (s *Server) handleAliveMember(member serf.Member) error {
 			ID:      structs.ConsulServiceID,
 			Service: structs.ConsulServiceName,
 			Port:    parts.Port,
+			Weights: &structs.Weights{
+				Passing: 1,
+				Warning: 1,
+			},
+			Meta: map[string]string{
+				"raft_version":          strconv.Itoa(parts.RaftVersion),
+				"serf_protocol_current": strconv.FormatUint(uint64(member.ProtocolCur), 10),
+				"serf_protocol_min":     strconv.FormatUint(uint64(member.ProtocolMin), 10),
+				"serf_protocol_max":     strconv.FormatUint(uint64(member.ProtocolMax), 10),
+				"version":               parts.Build.String(),
+			},
 		}
 
 		// Attempt to join the consul server
@@ -1317,11 +1169,12 @@ AFTER_CHECK:
 			Status:  api.HealthPassing,
 			Output:  structs.SerfCheckAliveOutput,
 		},
-
-		// If there's existing information about the node, do not
-		// clobber it.
-		SkipNodeUpdate: true,
 	}
+	if node != nil {
+		req.TaggedAddresses = node.TaggedAddresses
+		req.NodeMeta = node.Meta
+	}
+
 	_, err = s.raftApply(structs.RegisterRequestType, &req)
 	return err
 }
@@ -1335,7 +1188,13 @@ func (s *Server) handleFailedMember(member serf.Member) error {
 	if err != nil {
 		return err
 	}
-	if node != nil && node.Address == member.Addr.String() {
+
+	if node == nil {
+		s.logger.Printf("[INFO] consul: ignoring failed event for member '%s' because it does not exist in the catalog", member.Name)
+		return nil
+	}
+
+	if node.Address == member.Addr.String() {
 		// Check if the serfCheck is in the critical state
 		_, checks, err := state.NodeChecks(nil, member.Name)
 		if err != nil {
