@@ -2,7 +2,6 @@ package consul
 
 import (
 	"context"
-	"crypto/tls"
 	"errors"
 	"fmt"
 	"io"
@@ -14,10 +13,12 @@ import (
 	"path/filepath"
 	"reflect"
 	"strconv"
+	"strings"
 	"sync"
 	"sync/atomic"
 	"time"
 
+	metrics "github.com/armon/go-metrics"
 	ca "github.com/hashicorp/consul/agent/connect/ca"
 	"github.com/hashicorp/consul/agent/consul/autopilot"
 	"github.com/hashicorp/consul/agent/consul/fsm"
@@ -31,9 +32,12 @@ import (
 	"github.com/hashicorp/consul/sentinel"
 	"github.com/hashicorp/consul/tlsutil"
 	"github.com/hashicorp/consul/types"
+	"github.com/hashicorp/go-hclog"
+	"github.com/hashicorp/go-memdb"
 	"github.com/hashicorp/raft"
 	raftboltdb "github.com/hashicorp/raft-boltdb"
 	"github.com/hashicorp/serf/serf"
+	"golang.org/x/time/rate"
 )
 
 // These are the protocol versions that Consul can _understand_. These are
@@ -109,6 +113,15 @@ type Server struct {
 	aclReplicationLock    sync.RWMutex
 	aclReplicationEnabled bool
 
+	// aclTokenReapCancel is used to shut down the ACL Token expiration reap
+	// goroutine when we lose leadership.
+	aclTokenReapCancel  context.CancelFunc
+	aclTokenReapLock    sync.RWMutex
+	aclTokenReapEnabled bool
+
+	aclAuthMethodValidators    map[string]*authMethodValidatorEntry
+	aclAuthMethodValidatorLock sync.RWMutex
+
 	// DEPRECATED (ACL-Legacy-Compat) - only needed while we support both
 	// useNewACLs is used to determine whether we can use new ACLs or not
 	useNewACLs int32
@@ -129,15 +142,13 @@ type Server struct {
 	caProviderRoot *structs.CARoot
 	caProviderLock sync.RWMutex
 
-	// caPruningCh is used to shut down the CA root pruning goroutine when we
-	// lose leadership.
-	caPruningCh      chan struct{}
-	caPruningLock    sync.RWMutex
-	caPruningEnabled bool
-
 	// Consul configuration
 	config *Config
 
+	// configReplicator is used to manage the leaders replication routines for
+	// centralized config
+	configReplicator *Replicator
+
 	// tokens holds ACL tokens initially from the configuration, but can
 	// be updated at runtime, so should always be used instead of going to
 	// the configuration directly.
@@ -193,12 +204,24 @@ type Server struct {
 	// Enterprise user-defined areas.
 	router *router.Router
 
+	// rpcLimiter is used to rate limit the total number of RPCs initiated
+	// from an agent.
+	rpcLimiter atomic.Value
+
 	// Listener is used to listen for incoming connections
 	Listener  net.Listener
 	rpcServer *rpc.Server
 
-	// rpcTLS is the TLS config for incoming TLS requests
-	rpcTLS *tls.Config
+	// insecureRPCServer is a RPC server that is configure with
+	// IncomingInsecureRPCConfig to allow clients to call AutoEncrypt.Sign
+	// to request client certificates. At this point a client doesn't have
+	// a client cert and thus cannot present it. This is the only RPC
+	// Endpoint that is available at the time of writing.
+	insecureRPCServer *rpc.Server
+
+	// tlsConfigurator holds the agent configuration relevant to TLS and
+	// configures everything related to it.
+	tlsConfigurator *tlsutil.Configurator
 
 	// serfLAN is the Serf cluster maintained inside the DC
 	// which contains all the DC nodes
@@ -248,17 +271,32 @@ type Server struct {
 	shutdownCh   chan struct{}
 	shutdownLock sync.Mutex
 
+	// State for multi-dc connect leader logic
+	connectLock    sync.RWMutex
+	connectEnabled bool
+	connectCh      chan struct{}
+
+	// State for whether this datacenter is acting as a secondary CA.
+	actingSecondaryCA   bool
+	actingSecondaryLock sync.RWMutex
+
 	// embedded struct to hold all the enterprise specific data
 	EnterpriseServer
 }
 
+// NewServer is only used to help setting up a server for testing. Normal code
+// exercises NewServerLogger.
 func NewServer(config *Config) (*Server, error) {
-	return NewServerLogger(config, nil, new(token.Store))
+	c, err := tlsutil.NewConfigurator(config.ToTLSUtilConfig(), nil)
+	if err != nil {
+		return nil, err
+	}
+	return NewServerLogger(config, nil, new(token.Store), c)
 }
 
-// NewServer is used to construct a new Consul server from the
+// NewServerLogger is used to construct a new Consul server from the
 // configuration, potentially returning an error
-func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*Server, error) {
+func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store, tlsConfigurator *tlsutil.Configurator) (*Server, error) {
 	// Check the protocol version.
 	if err := config.CheckProtocolVersion(); err != nil {
 		return nil, err
@@ -296,19 +334,6 @@ func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*
 		}
 	}
 
-	// Create the TLS wrapper for outgoing connections.
-	tlsConf := config.tlsConfig()
-	tlsWrap, err := tlsConf.OutgoingTLSWrapper()
-	if err != nil {
-		return nil, err
-	}
-
-	// Get the incoming TLS config.
-	incomingTLS, err := tlsConf.IncomingTLSConfig()
-	if err != nil {
-		return nil, err
-	}
-
 	// Create the tombstone GC.
 	gc, err := state.NewTombstoneGC(config.TombstoneTTL, config.TombstoneTTLGranularity)
 	if err != nil {
@@ -319,33 +344,34 @@ func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*
 	shutdownCh := make(chan struct{})
 
 	connPool := &pool.ConnPool{
-		SrcAddr:    config.RPCSrcAddr,
-		LogOutput:  config.LogOutput,
-		MaxTime:    serverRPCCache,
-		MaxStreams: serverMaxStreams,
-		TLSWrapper: tlsWrap,
-		ForceTLS:   config.VerifyOutgoing,
+		SrcAddr:         config.RPCSrcAddr,
+		LogOutput:       config.LogOutput,
+		MaxTime:         serverRPCCache,
+		MaxStreams:      serverMaxStreams,
+		TLSConfigurator: tlsConfigurator,
+		ForceTLS:        config.VerifyOutgoing,
 	}
 
 	// Create server.
 	s := &Server{
-		config:           config,
-		tokens:           tokens,
-		connPool:         connPool,
-		eventChLAN:       make(chan serf.Event, serfEventChSize),
-		eventChWAN:       make(chan serf.Event, serfEventChSize),
-		logger:           logger,
-		leaveCh:          make(chan struct{}),
-		reconcileCh:      make(chan serf.Member, reconcileChSize),
-		router:           router.NewRouter(logger, config.Datacenter),
-		rpcServer:        rpc.NewServer(),
-		rpcTLS:           incomingTLS,
-		reassertLeaderCh: make(chan chan error),
-		segmentLAN:       make(map[string]*serf.Serf, len(config.Segments)),
-		sessionTimers:    NewSessionTimers(),
-		tombstoneGC:      gc,
-		serverLookup:     NewServerLookup(),
-		shutdownCh:       shutdownCh,
+		config:            config,
+		tokens:            tokens,
+		connPool:          connPool,
+		eventChLAN:        make(chan serf.Event, serfEventChSize),
+		eventChWAN:        make(chan serf.Event, serfEventChSize),
+		logger:            logger,
+		leaveCh:           make(chan struct{}),
+		reconcileCh:       make(chan serf.Member, reconcileChSize),
+		router:            router.NewRouter(logger, config.Datacenter),
+		rpcServer:         rpc.NewServer(),
+		insecureRPCServer: rpc.NewServer(),
+		tlsConfigurator:   tlsConfigurator,
+		reassertLeaderCh:  make(chan chan error),
+		segmentLAN:        make(map[string]*serf.Serf, len(config.Segments)),
+		sessionTimers:     NewSessionTimers(),
+		tombstoneGC:       gc,
+		serverLookup:      NewServerLookup(),
+		shutdownCh:        shutdownCh,
 	}
 
 	// Initialize enterprise specific server functionality
@@ -354,6 +380,21 @@ func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*
 		return nil, err
 	}
 
+	s.rpcLimiter.Store(rate.NewLimiter(config.RPCRate, config.RPCMaxBurst))
+
+	configReplicatorConfig := ReplicatorConfig{
+		Name:        "Config Entry",
+		ReplicateFn: s.replicateConfig,
+		Rate:        s.config.ConfigReplicationRate,
+		Burst:       s.config.ConfigReplicationBurst,
+		Logger:      logger,
+	}
+	s.configReplicator, err = NewReplicator(&configReplicatorConfig)
+	if err != nil {
+		s.Shutdown()
+		return nil, err
+	}
+
 	// Initialize the stats fetcher that autopilot will use.
 	s.statsFetcher = NewStatsFetcher(logger, s.connPool, s.config.Datacenter)
 
@@ -374,7 +415,7 @@ func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*
 	}
 
 	// Initialize the RPC layer.
-	if err := s.setupRPC(tlsWrap); err != nil {
+	if err := s.setupRPC(); err != nil {
 		s.Shutdown()
 		return nil, fmt.Errorf("Failed to start RPC layer: %v", err)
 	}
@@ -392,6 +433,10 @@ func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*
 		return nil, fmt.Errorf("Failed to start Raft: %v", err)
 	}
 
+	if s.config.ConnectEnabled && s.config.AutoEncryptAllowTLS {
+		go s.trackAutoEncryptCARoots()
+	}
+
 	// Serf and dynamic bind ports
 	//
 	// The LAN serf cluster announces the port of the WAN serf cluster
@@ -462,6 +507,10 @@ func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*
 		return nil, err
 	}
 
+	// Initialize Autopilot. This must happen before starting leadership monitoring
+	// as establishing leadership could attempt to use autopilot and cause a panic.
+	s.initAutopilot(config)
+
 	// Start monitoring leadership. This must happen after Serf is set up
 	// since it can fire events when leadership is obtained.
 	go s.monitorLeadership()
@@ -477,12 +526,36 @@ func NewServerLogger(config *Config, logger *log.Logger, tokens *token.Store) (*
 	// Start the metrics handlers.
 	go s.sessionStats()
 
-	// Initialize Autopilot
-	s.initAutopilot(config)
-
 	return s, nil
 }
 
+func (s *Server) trackAutoEncryptCARoots() {
+	for {
+		select {
+		case <-s.shutdownCh:
+			s.logger.Printf("[DEBUG] agent: shutting down trackAutoEncryptCARoots because shutdown")
+			return
+		default:
+		}
+		ws := memdb.NewWatchSet()
+		state := s.fsm.State()
+		ws.Add(state.AbandonCh())
+		_, cas, err := state.CARoots(ws)
+		if err != nil {
+			s.logger.Printf("[DEBUG] agent: Failed to watch AutoEncrypt CARoot: %v", err)
+			return
+		}
+		caPems := []string{}
+		for _, ca := range cas {
+			caPems = append(caPems, ca.RootCert)
+		}
+		if err := s.tlsConfigurator.UpdateAutoEncryptCA(caPems); err != nil {
+			s.logger.Printf("[DEBUG] agent: Failed to update AutoEncrypt CAPems: %v", err)
+		}
+		ws.Watch(nil)
+	}
+}
+
 // setupRaft is used to setup and initialize Raft
 func (s *Server) setupRaft() error {
 	// If we have an unclean exit then attempt to close the Raft store.
@@ -512,6 +585,7 @@ func (s *Server) setupRaft() error {
 		MaxPool:               3,
 		Timeout:               10 * time.Second,
 		ServerAddressProvider: serverAddressProvider,
+		Logger:                s.logger,
 	}
 
 	trans := raft.NewNetworkTransportWithConfig(transConfig)
@@ -519,7 +593,13 @@ func (s *Server) setupRaft() error {
 
 	// Make sure we set the LogOutput.
 	s.config.RaftConfig.LogOutput = s.config.LogOutput
-	s.config.RaftConfig.Logger = s.logger
+	raftLogger := hclog.New(&hclog.LoggerOptions{
+		Name:       "raft",
+		Level:      hclog.LevelFromString(s.config.LogLevel),
+		Output:     s.config.LogOutput,
+		TimeFormat: `2006/01/02 15:04:05`,
+	})
+	s.config.RaftConfig.Logger = raftLogger
 
 	// Versions of the Raft protocol below 3 require the LocalID to match the network
 	// address of the transport.
@@ -648,7 +728,7 @@ func (s *Server) setupRaft() error {
 	s.raftNotifyCh = raftNotifyCh
 
 	// Setup the Raft store.
-	s.raft, err = raft.NewRaft(s.config.RaftConfig, s.fsm, log, stable, snap, trans)
+	s.raft, err = raft.NewRaft(s.config.RaftConfig, s.fsm.ChunkingFSM(), log, stable, snap, trans)
 	if err != nil {
 		return err
 	}
@@ -668,11 +748,16 @@ func registerEndpoint(fn factory) {
 }
 
 // setupRPC is used to setup the RPC listener
-func (s *Server) setupRPC(tlsWrap tlsutil.DCWrapper) error {
+func (s *Server) setupRPC() error {
 	for _, fn := range endpoints {
 		s.rpcServer.Register(fn(s))
 	}
 
+	// Only register AutoEncrypt on the insecure RPC server. Insecure only
+	// means that verify incoming is turned off even though it might have
+	// been configured.
+	s.insecureRPCServer.Register(&AutoEncrypt{srv: s})
+
 	ln, err := net.ListenTCP("tcp", s.config.RPCAddr)
 	if err != nil {
 		return err
@@ -694,7 +779,7 @@ func (s *Server) setupRPC(tlsWrap tlsutil.DCWrapper) error {
 
 	// Provide a DC specific wrapper. Raft replication is only
 	// ever done in the same datacenter, so we can provide it as a constant.
-	wrapper := tlsutil.SpecificDC(s.config.Datacenter, tlsWrap)
+	wrapper := tlsutil.SpecificDC(s.config.Datacenter, s.tlsConfigurator.OutgoingRPCWrapper())
 
 	// Define a callback for determining whether to wrap a connection with TLS
 	tlsFunc := func(address raft.ServerAddress) bool {
@@ -812,6 +897,9 @@ func (s *Server) Leave() error {
 		}
 	}
 
+	// Leave everything enterprise related as well
+	s.handleEnterpriseLeave()
+
 	// Start refusing RPCs now that we've left the LAN pool. It's important
 	// to do this *after* we've left the LAN pool so that clients will know
 	// to shift onto another server if they perform a retry. We also wake up
@@ -922,6 +1010,11 @@ func (s *Server) RemoveFailedNode(node string) error {
 	if err := s.serfLAN.RemoveFailedNode(node); err != nil {
 		return err
 	}
+	// The Serf WAN pool stores members as node.datacenter
+	// so the dc is appended if not present
+	if !strings.HasSuffix(node, "."+s.config.Datacenter) {
+		node = node + "." + s.config.Datacenter
+	}
 	if s.serfWAN != nil {
 		if err := s.serfWAN.RemoveFailedNode(node); err != nil {
 			return err
@@ -1007,6 +1100,19 @@ func (s *Server) RPC(method string, args interface{}, reply interface{}) error {
 		args:   args,
 		reply:  reply,
 	}
+
+	// Enforce the RPC limit.
+	//
+	// "client" metric path because the internal client API is calling to the
+	// internal server API. It's odd that the same request directed to a server is
+	// recorded differently. On the other hand this possibly masks the different
+	// between regular client requests that traverse the network and these which
+	// don't (unless forwarded). This still seems most sane.
+	metrics.IncrCounter([]string{"client", "rpc"}, 1)
+	if !s.rpcLimiter.Load().(*rate.Limiter).Allow() {
+		metrics.IncrCounter([]string{"client", "rpc", "exceeded"}, 1)
+		return structs.ErrRPCRateExceeded
+	}
 	if err := s.rpcServer.ServeRequest(codec); err != nil {
 		return err
 	}
@@ -1018,6 +1124,19 @@ func (s *Server) RPC(method string, args interface{}, reply interface{}) error {
 func (s *Server) SnapshotRPC(args *structs.SnapshotRequest, in io.Reader, out io.Writer,
 	replyFn structs.SnapshotReplyFn) error {
 
+	// Enforce the RPC limit.
+	//
+	// "client" metric path because the internal client API is calling to the
+	// internal server API. It's odd that the same request directed to a server is
+	// recorded differently. On the other hand this possibly masks the different
+	// between regular client requests that traverse the network and these which
+	// don't (unless forwarded). This still seems most sane.
+	metrics.IncrCounter([]string{"client", "rpc"}, 1)
+	if !s.rpcLimiter.Load().(*rate.Limiter).Allow() {
+		metrics.IncrCounter([]string{"client", "rpc", "exceeded"}, 1)
+		return structs.ErrRPCRateExceeded
+	}
+
 	// Perform the operation.
 	var reply structs.SnapshotResponse
 	snap, err := s.dispatchSnapshotRequest(args, in, &reply)
@@ -1120,6 +1239,13 @@ func (s *Server) GetLANCoordinate() (lib.CoordinateSet, error) {
 // ReloadConfig is used to have the Server do an online reload of
 // relevant configuration information
 func (s *Server) ReloadConfig(config *Config) error {
+	s.rpcLimiter.Store(rate.NewLimiter(config.RPCRate, config.RPCMaxBurst))
+
+	if s.IsLeader() {
+		// only bootstrap the config entries if we are the leader
+		// this will error if we lose leadership while bootstrapping here.
+		return s.bootstrapConfigEntries(config.ConfigEntryBootstrap)
+	}
 	return nil
 }
 
@@ -1138,6 +1264,10 @@ func (s *Server) isReadyForConsistentReads() bool {
 	return atomic.LoadInt32(&s.readyForConsistentReads) == 1
 }
 
+func (s *Server) intentionReplicationEnabled() bool {
+	return s.config.ConnectEnabled && s.config.Datacenter != s.config.PrimaryDatacenter
+}
+
 // peersInfoContent is used to help operators understand what happened to the
 // peers.json file. This is written to a file called peers.info in the same
 // location.
