@@ -3,7 +3,6 @@ package local
 import (
 	"fmt"
 	"log"
-	"math/rand"
 	"reflect"
 	"strconv"
 	"strings"
@@ -11,7 +10,7 @@ import (
 	"sync/atomic"
 	"time"
 
-	"github.com/armon/go-metrics"
+	metrics "github.com/armon/go-metrics"
 
 	"github.com/hashicorp/consul/acl"
 	"github.com/hashicorp/consul/agent/structs"
@@ -19,9 +18,10 @@ import (
 	"github.com/hashicorp/consul/api"
 	"github.com/hashicorp/consul/lib"
 	"github.com/hashicorp/consul/types"
-	"github.com/hashicorp/go-uuid"
 )
 
+const fullSyncReadMaxStale = 2 * time.Second
+
 // Config is the configuration for the State.
 type Config struct {
 	AdvertiseAddr       string
@@ -31,8 +31,6 @@ type Config struct {
 	NodeID              types.NodeID
 	NodeName            string
 	TaggedAddresses     map[string]string
-	ProxyBindMinPort    int
-	ProxyBindMaxPort    int
 }
 
 // ServiceState describes the state of a service record.
@@ -70,6 +68,9 @@ func (s *ServiceState) Clone() *ServiceState {
 // CheckState describes the state of a health check record.
 type CheckState struct {
 	// Check is the local copy of the health check record.
+	//
+	// Must Clone() the overall CheckState before mutating this. After mutation
+	// reinstall into the checks map.
 	Check *structs.HealthCheck
 
 	// Token is the ACL record to update or delete the health check
@@ -95,12 +96,15 @@ type CheckState struct {
 	Deleted bool
 }
 
-// Clone returns a shallow copy of the object. The check record and the
-// defer timer still point to the original values and must not be
-// modified.
+// Clone returns a shallow copy of the object.
+//
+// The defer timer still points to the original value and must not be modified.
 func (c *CheckState) Clone() *CheckState {
 	c2 := new(CheckState)
 	*c2 = *c
+	if c.Check != nil {
+		c2.Check = c.Check.Clone()
+	}
 	return c2
 }
 
@@ -119,32 +123,6 @@ type rpc interface {
 	RPC(method string, args interface{}, reply interface{}) error
 }
 
-// ManagedProxy represents the local state for a registered proxy instance.
-type ManagedProxy struct {
-	Proxy *structs.ConnectManagedProxy
-
-	// ProxyToken is a special local-only security token that grants the bearer
-	// access to the proxy's config as well as allowing it to request certificates
-	// on behalf of the target service. Certain connect endpoints will validate
-	// against this token and if it matches will then use the target service's
-	// registration token to actually authenticate the upstream RPC on behalf of
-	// the service. This token is passed securely to the proxy process via ENV
-	// vars and should never be exposed any other way. Unmanaged proxies will
-	// never see this and need to use service-scoped ACL tokens distributed
-	// externally. It is persisted in the local state to allow authenticating
-	// running proxies after the agent restarts.
-	//
-	// TODO(banks): In theory we only need to persist this at all to _validate_
-	// which means we could keep only a hash in memory and on disk and only pass
-	// the actual token to the process on startup. That would require a bit of
-	// refactoring though to have the required interaction with the proxy manager.
-	ProxyToken string
-
-	// WatchCh is a close-only chan that is closed when the proxy is removed or
-	// updated.
-	WatchCh chan struct{}
-}
-
 // State is used to represent the node's services,
 // and checks. We use it to perform anti-entropy with the
 // catalog representation
@@ -193,42 +171,22 @@ type State struct {
 	// notifyHandlers is a map of registered channel listeners that are sent
 	// messages whenever state changes occur. For now these events only include
 	// service registration and deregistration since that is all that is needed
-	// but the same mechanism could be used for other state changes.
-	//
-	// Note that we haven't refactored managedProxyHandlers into this mechanism
-	// yet because that is soon to be deprecated and removed so it's easier to
-	// just leave them separate until managed proxies are removed entirely. Any
-	// future notifications should re-use this mechanism though.
+	// but the same mechanism could be used for other state changes. Any
+	// future notifications should re-use this mechanism.
 	notifyHandlers map[chan<- struct{}]struct{}
-
-	// managedProxies is a map of all managed connect proxies registered locally on
-	// this agent. This is NOT kept in sync with servers since it's agent-local
-	// config only. Proxy instances have separate service registrations in the
-	// services map above which are kept in sync via anti-entropy. Un-managed
-	// proxies (that registered themselves separately from the service
-	// registration) do not appear here as the agent doesn't need to manage their
-	// process nor config. The _do_ still exist in services above though as
-	// services with Kind == connect-proxy.
-	//
-	// managedProxyHandlers is a map of registered channel listeners that
-	// are sent a message each time a proxy changes via Add or RemoveProxy.
-	managedProxies       map[string]*ManagedProxy
-	managedProxyHandlers map[chan<- struct{}]struct{}
 }
 
 // NewState creates a new local state for the agent.
 func NewState(c Config, lg *log.Logger, tokens *token.Store) *State {
 	l := &State{
-		config:               c,
-		logger:               lg,
-		services:             make(map[string]*ServiceState),
-		checks:               make(map[types.CheckID]*CheckState),
-		checkAliases:         make(map[string]map[types.CheckID]chan<- struct{}),
-		metadata:             make(map[string]string),
-		tokens:               tokens,
-		notifyHandlers:       make(map[chan<- struct{}]struct{}),
-		managedProxies:       make(map[string]*ManagedProxy),
-		managedProxyHandlers: make(map[chan<- struct{}]struct{}),
+		config:         c,
+		logger:         lg,
+		services:       make(map[string]*ServiceState),
+		checks:         make(map[types.CheckID]*CheckState),
+		checkAliases:   make(map[string]map[types.CheckID]chan<- struct{}),
+		metadata:       make(map[string]string),
+		tokens:         tokens,
+		notifyHandlers: make(map[chan<- struct{}]struct{}),
 	}
 	l.SetDiscardCheckOutput(c.DiscardCheckOutput)
 	return l
@@ -265,6 +223,12 @@ func (l *State) serviceToken(id string) string {
 // This entry is persistent and the agent will make a best effort to
 // ensure it is registered
 func (l *State) AddService(service *structs.NodeService, token string) error {
+	l.Lock()
+	defer l.Unlock()
+	return l.addServiceLocked(service, token)
+}
+
+func (l *State) addServiceLocked(service *structs.NodeService, token string) error {
 	if service == nil {
 		return fmt.Errorf("no service")
 	}
@@ -274,18 +238,58 @@ func (l *State) AddService(service *structs.NodeService, token string) error {
 		service.ID = service.Service
 	}
 
-	l.SetServiceState(&ServiceState{
+	l.setServiceStateLocked(&ServiceState{
 		Service: service,
 		Token:   token,
 	})
 	return nil
 }
 
+// AddServiceWithChecks adds a service and its check tp the local state atomically
+func (l *State) AddServiceWithChecks(service *structs.NodeService, checks []*structs.HealthCheck, token string) error {
+	l.Lock()
+	defer l.Unlock()
+
+	if err := l.addServiceLocked(service, token); err != nil {
+		return err
+	}
+
+	for _, check := range checks {
+		if err := l.addCheckLocked(check, token); err != nil {
+			return err
+		}
+	}
+
+	return nil
+}
+
 // RemoveService is used to remove a service entry from the local state.
 // The agent will make a best effort to ensure it is deregistered.
 func (l *State) RemoveService(id string) error {
 	l.Lock()
 	defer l.Unlock()
+	return l.removeServiceLocked(id)
+}
+
+// RemoveServiceWithChecks removes a service and its check from the local state atomically
+func (l *State) RemoveServiceWithChecks(serviceID string, checkIDs []types.CheckID) error {
+	l.Lock()
+	defer l.Unlock()
+
+	if err := l.removeServiceLocked(serviceID); err != nil {
+		return err
+	}
+
+	for _, id := range checkIDs {
+		if err := l.removeCheckLocked(id); err != nil {
+			return err
+		}
+	}
+
+	return nil
+}
+
+func (l *State) removeServiceLocked(id string) error {
 
 	s := l.services[id]
 	if s == nil || s.Deleted {
@@ -358,6 +362,10 @@ func (l *State) SetServiceState(s *ServiceState) {
 	l.Lock()
 	defer l.Unlock()
 
+	l.setServiceStateLocked(s)
+}
+
+func (l *State) setServiceStateLocked(s *ServiceState) {
 	s.WatchCh = make(chan struct{})
 
 	old, hasOld := l.services[s.Service.ID]
@@ -414,6 +422,13 @@ func (l *State) checkToken(id types.CheckID) string {
 // This entry is persistent and the agent will make a best effort to
 // ensure it is registered
 func (l *State) AddCheck(check *structs.HealthCheck, token string) error {
+	l.Lock()
+	defer l.Unlock()
+
+	return l.addCheckLocked(check, token)
+}
+
+func (l *State) addCheckLocked(check *structs.HealthCheck, token string) error {
 	if check == nil {
 		return fmt.Errorf("no check")
 	}
@@ -427,14 +442,14 @@ func (l *State) AddCheck(check *structs.HealthCheck, token string) error {
 
 	// if there is a serviceID associated with the check, make sure it exists before adding it
 	// NOTE - This logic may be moved to be handled within the Agent's Addcheck method after a refactor
-	if check.ServiceID != "" && l.Service(check.ServiceID) == nil {
+	if _, ok := l.services[check.ServiceID]; check.ServiceID != "" && !ok {
 		return fmt.Errorf("Check %q refers to non-existent service %q", check.CheckID, check.ServiceID)
 	}
 
 	// hard-set the node name
 	check.Node = l.config.NodeName
 
-	l.SetCheckState(&CheckState{
+	l.setCheckStateLocked(&CheckState{
 		Check: check,
 		Token: token,
 	})
@@ -477,17 +492,23 @@ func (l *State) RemoveAliasCheck(checkID types.CheckID, srcServiceID string) {
 
 // RemoveCheck is used to remove a health check from the local state.
 // The agent will make a best effort to ensure it is deregistered
-// todo(fs): RemoveService returns an error for a non-existant service. RemoveCheck should as well.
+// todo(fs): RemoveService returns an error for a non-existent service. RemoveCheck should as well.
 // todo(fs): Check code that calls this to handle the error.
 func (l *State) RemoveCheck(id types.CheckID) error {
 	l.Lock()
 	defer l.Unlock()
+	return l.removeCheckLocked(id)
+}
 
+func (l *State) removeCheckLocked(id types.CheckID) error {
 	c := l.checks[id]
 	if c == nil || c.Deleted {
 		return fmt.Errorf("Check %q does not exist", id)
 	}
 
+	// If this is a check for an aliased service, then notify the waiters.
+	l.notifyIfAliased(c.Check.ServiceID)
+
 	// To remove the check on the server we need the token.
 	// Therefore, we mark the service as deleted and keep the
 	// entry around until it is actually removed.
@@ -527,6 +548,18 @@ func (l *State) UpdateCheck(id types.CheckID, status, output string) {
 		return
 	}
 
+	// Ensure we only mutate a copy of the check state and put the finalized
+	// version into the checks map when complete.
+	//
+	// Note that we are relying upon the earlier deferred mutex unlock to
+	// happen AFTER this defer. As per the Go spec this is true, but leaving
+	// this note here for the future in case of any refactorings which may not
+	// notice this relationship.
+	c = c.Clone()
+	defer func(c *CheckState) {
+		l.checks[id] = c
+	}(c)
+
 	// Defer a sync if the output has changed. This is an optimization around
 	// frequent updates of output. Instead, we update the output internally,
 	// and periodically do a write-back to the servers. If there is a status
@@ -556,18 +589,7 @@ func (l *State) UpdateCheck(id types.CheckID, status, output string) {
 	}
 
 	// If this is a check for an aliased service, then notify the waiters.
-	if aliases, ok := l.checkAliases[c.Check.ServiceID]; ok && len(aliases) > 0 {
-		for _, notifyCh := range aliases {
-			// Do not block. All notify channels should be buffered to at
-			// least 1 in which case not-blocking does not result in loss
-			// of data because a failed send means a notification is
-			// already queued. This must be called with the lock held.
-			select {
-			case notifyCh <- struct{}{}:
-			default:
-			}
-		}
-	}
+	l.notifyIfAliased(c.Check.ServiceID)
 
 	// Update status and mark out of sync
 	c.Check.Status = status
@@ -599,9 +621,9 @@ func (l *State) Checks() map[types.CheckID]*structs.HealthCheck {
 	return m
 }
 
-// CheckState returns a shallow copy of the current health check state
-// record. The health check record and the deferred check still point to
-// the original values and must not be modified.
+// CheckState returns a shallow copy of the current health check state record.
+//
+// The defer timer still points to the original value and must not be modified.
 func (l *State) CheckState(id types.CheckID) *CheckState {
 	l.RLock()
 	defer l.RUnlock()
@@ -620,13 +642,22 @@ func (l *State) SetCheckState(c *CheckState) {
 	l.Lock()
 	defer l.Unlock()
 
+	l.setCheckStateLocked(c)
+}
+
+func (l *State) setCheckStateLocked(c *CheckState) {
 	l.checks[c.Check.CheckID] = c
+
+	// If this is a check for an aliased service, then notify the waiters.
+	l.notifyIfAliased(c.Check.ServiceID)
+
 	l.TriggerSyncChanges()
 }
 
 // CheckStates returns a shallow copy of all health check state records.
-// The health check records and the deferred checks still point to
-// the original values and must not be modified.
+// The map contains a shallow copy of the current check states.
+//
+// The defer timers still point to the original values and must not be modified.
 func (l *State) CheckStates() map[types.CheckID]*CheckState {
 	l.RLock()
 	defer l.RUnlock()
@@ -643,9 +674,9 @@ func (l *State) CheckStates() map[types.CheckID]*CheckState {
 
 // CriticalCheckStates returns the locally registered checks that the
 // agent is aware of and are being kept in sync with the server.
-// The map contains a shallow copy of the current check states but
-// references to the actual check definition which must not be
-// modified.
+// The map contains a shallow copy of the current check states.
+//
+// The defer timers still point to the original values and must not be modified.
 func (l *State) CriticalCheckStates() map[types.CheckID]*CheckState {
 	l.RLock()
 	defer l.RUnlock()
@@ -660,188 +691,6 @@ func (l *State) CriticalCheckStates() map[types.CheckID]*CheckState {
 	return m
 }
 
-// AddProxy is used to add a connect proxy entry to the local state. This
-// assumes the proxy's NodeService is already registered via Agent.AddService
-// (since that has to do other book keeping). The token passed here is the ACL
-// token the service used to register itself so must have write on service
-// record. AddProxy returns the newly added proxy and an error.
-//
-// The restoredProxyToken argument should only be used when restoring proxy
-// definitions from disk; new proxies must leave it blank to get a new token
-// assigned. We need to restore from disk to enable to continue authenticating
-// running proxies that already had that credential injected.
-func (l *State) AddProxy(proxy *structs.ConnectManagedProxy, token,
-	restoredProxyToken string) (*ManagedProxy, error) {
-	if proxy == nil {
-		return nil, fmt.Errorf("no proxy")
-	}
-
-	// Lookup the local service
-	target := l.Service(proxy.TargetServiceID)
-	if target == nil {
-		return nil, fmt.Errorf("target service ID %s not registered",
-			proxy.TargetServiceID)
-	}
-
-	// Get bind info from config
-	cfg, err := proxy.ParseConfig()
-	if err != nil {
-		return nil, err
-	}
-
-	// Construct almost all of the NodeService that needs to be registered by the
-	// caller outside of the lock.
-	svc := &structs.NodeService{
-		Kind:    structs.ServiceKindConnectProxy,
-		ID:      target.ID + "-proxy",
-		Service: target.Service + "-proxy",
-		Proxy: structs.ConnectProxyConfig{
-			DestinationServiceName: target.Service,
-			LocalServiceAddress:    cfg.LocalServiceAddress,
-			LocalServicePort:       cfg.LocalServicePort,
-		},
-		Address: cfg.BindAddress,
-		Port:    cfg.BindPort,
-	}
-
-	// Set default port now while the target is known
-	if svc.Proxy.LocalServicePort < 1 {
-		svc.Proxy.LocalServicePort = target.Port
-	}
-
-	// Lock now. We can't lock earlier as l.Service would deadlock and shouldn't
-	// anyway to minimise the critical section.
-	l.Lock()
-	defer l.Unlock()
-
-	pToken := restoredProxyToken
-
-	// Does this proxy instance allready exist?
-	if existing, ok := l.managedProxies[svc.ID]; ok {
-		// Keep the existing proxy token so we don't have to restart proxy to
-		// re-inject token.
-		pToken = existing.ProxyToken
-		// If the user didn't explicitly change the port, use the old one instead of
-		// assigning new.
-		if svc.Port < 1 {
-			svc.Port = existing.Proxy.ProxyService.Port
-		}
-	} else if proxyService, ok := l.services[svc.ID]; ok {
-		// The proxy-service already exists so keep the port that got assigned. This
-		// happens on reload from disk since service definitions are reloaded first.
-		svc.Port = proxyService.Service.Port
-	}
-
-	// If this is a new instance, generate a token
-	if pToken == "" {
-		pToken, err = uuid.GenerateUUID()
-		if err != nil {
-			return nil, err
-		}
-	}
-
-	// Allocate port if needed (min and max inclusive).
-	rangeLen := l.config.ProxyBindMaxPort - l.config.ProxyBindMinPort + 1
-	if svc.Port < 1 && l.config.ProxyBindMinPort > 0 && rangeLen > 0 {
-		// This should be a really short list so don't bother optimising lookup yet.
-	OUTER:
-		for _, offset := range rand.Perm(rangeLen) {
-			p := l.config.ProxyBindMinPort + offset
-			// See if this port was already allocated to another proxy
-			for _, other := range l.managedProxies {
-				if other.Proxy.ProxyService.Port == p {
-					// allready taken, skip to next random pick in the range
-					continue OUTER
-				}
-			}
-			// We made it through all existing proxies without a match so claim this one
-			svc.Port = p
-			break
-		}
-	}
-	// If no ports left (or auto ports disabled) fail
-	if svc.Port < 1 {
-		return nil, fmt.Errorf("no port provided for proxy bind_port and none "+
-			" left in the allocated range [%d, %d]", l.config.ProxyBindMinPort,
-			l.config.ProxyBindMaxPort)
-	}
-
-	proxy.ProxyService = svc
-
-	// All set, add the proxy and return the service
-	if old, ok := l.managedProxies[svc.ID]; ok {
-		// Notify watchers of the existing proxy config that it's changing. Note
-		// this is safe here even before the map is updated since we still hold the
-		// state lock and the watcher can't re-read the new config until we return
-		// anyway.
-		close(old.WatchCh)
-	}
-	l.managedProxies[svc.ID] = &ManagedProxy{
-		Proxy:      proxy,
-		ProxyToken: pToken,
-		WatchCh:    make(chan struct{}),
-	}
-
-	// Notify
-	for ch := range l.managedProxyHandlers {
-		// Do not block
-		select {
-		case ch <- struct{}{}:
-		default:
-		}
-	}
-
-	// No need to trigger sync as proxy state is local only.
-	return l.managedProxies[svc.ID], nil
-}
-
-// RemoveProxy is used to remove a proxy entry from the local state.
-// This returns the proxy that was removed.
-func (l *State) RemoveProxy(id string) (*ManagedProxy, error) {
-	l.Lock()
-	defer l.Unlock()
-
-	p := l.managedProxies[id]
-	if p == nil {
-		return nil, fmt.Errorf("Proxy %s does not exist", id)
-	}
-	delete(l.managedProxies, id)
-
-	// Notify watchers of the existing proxy config that it's changed.
-	close(p.WatchCh)
-
-	// Notify
-	for ch := range l.managedProxyHandlers {
-		// Do not block
-		select {
-		case ch <- struct{}{}:
-		default:
-		}
-	}
-
-	// No need to trigger sync as proxy state is local only.
-	return p, nil
-}
-
-// Proxy returns the local proxy state.
-func (l *State) Proxy(id string) *ManagedProxy {
-	l.RLock()
-	defer l.RUnlock()
-	return l.managedProxies[id]
-}
-
-// Proxies returns the locally registered proxies.
-func (l *State) Proxies() map[string]*ManagedProxy {
-	l.RLock()
-	defer l.RUnlock()
-
-	m := make(map[string]*ManagedProxy)
-	for id, p := range l.managedProxies {
-		m[id] = p
-	}
-	return m
-}
-
 // broadcastUpdateLocked assumes l is locked and delivers an update to all
 // registered watchers.
 func (l *State) broadcastUpdateLocked() {
@@ -877,31 +726,6 @@ func (l *State) StopNotify(ch chan<- struct{}) {
 	delete(l.notifyHandlers, ch)
 }
 
-// NotifyProxy will register a channel to receive messages when the
-// configuration or set of proxies changes. This will not block on
-// channel send so ensure the channel has a buffer. Note that any buffer
-// size is generally fine since actual data is not sent over the channel,
-// so a dropped send due to a full buffer does not result in any loss of
-// data. The fact that a buffer already contains a notification means that
-// the receiver will still be notified that changes occurred.
-//
-// NOTE(mitchellh): This could be more generalized but for my use case I
-// only needed proxy events. In the future if it were to be generalized I
-// would add a new Notify method and remove the proxy-specific ones.
-func (l *State) NotifyProxy(ch chan<- struct{}) {
-	l.Lock()
-	defer l.Unlock()
-	l.managedProxyHandlers[ch] = struct{}{}
-}
-
-// StopNotifyProxy will deregister a channel receiving proxy notifications.
-// Pair this with all calls to NotifyProxy to clean up state.
-func (l *State) StopNotifyProxy(ch chan<- struct{}) {
-	l.Lock()
-	defer l.Unlock()
-	delete(l.managedProxyHandlers, ch)
-}
-
 // Metadata returns the local node metadata fields that the
 // agent is aware of and are being kept in sync with the server
 func (l *State) Metadata() map[string]string {
@@ -967,9 +791,13 @@ func (l *State) Stats() map[string]string {
 func (l *State) updateSyncState() error {
 	// Get all checks and services from the master
 	req := structs.NodeSpecificRequest{
-		Datacenter:   l.config.Datacenter,
-		Node:         l.config.NodeName,
-		QueryOptions: structs.QueryOptions{Token: l.tokens.AgentToken()},
+		Datacenter: l.config.Datacenter,
+		Node:       l.config.NodeName,
+		QueryOptions: structs.QueryOptions{
+			Token:            l.tokens.AgentToken(),
+			AllowStale:       true,
+			MaxStaleDuration: fullSyncReadMaxStale,
+		},
 	}
 
 	var out1 structs.IndexedNodeServices
@@ -1410,3 +1238,19 @@ func (l *State) syncNodeInfo() error {
 		return err
 	}
 }
+
+// notifyIfAliased will notify waiters if this is a check for an aliased service
+func (l *State) notifyIfAliased(serviceID string) {
+	if aliases, ok := l.checkAliases[serviceID]; ok && len(aliases) > 0 {
+		for _, notifyCh := range aliases {
+			// Do not block. All notify channels should be buffered to at
+			// least 1 in which case not-blocking does not result in loss
+			// of data because a failed send means a notification is
+			// already queued. This must be called with the lock held.
+			select {
+			case notifyCh <- struct{}{}:
+			default:
+			}
+		}
+	}
+}
