@@ -1,20 +1,17 @@
 package xds
 
 import (
-	"bytes"
 	"context"
 	"errors"
-	"fmt"
 	"log"
 	"os"
 	"strings"
 	"sync"
+	"sync/atomic"
 	"testing"
-	"text/template"
 	"time"
 
 	envoy "github.com/envoyproxy/go-control-plane/envoy/api/v2"
-	"github.com/gogo/protobuf/jsonpb"
 	"github.com/stretchr/testify/require"
 	"google.golang.org/grpc/codes"
 	"google.golang.org/grpc/metadata"
@@ -82,7 +79,7 @@ func (m *testManager) Watch(proxyID string) (<-chan *proxycfg.ConfigSnapshot, pr
 // AssertWatchCancelled checks that the most recent call to a Watch cancel func
 // was from the specified proxyID and that one is made in a short time. This
 // probably won't work if you are running multiple Watches in parallel on
-// multiple proxyIDS due to timing/ordering issues but I dont think we need to
+// multiple proxyIDS due to timing/ordering issues but I don't think we need to
 // do that.
 func (m *testManager) AssertWatchCancelled(t *testing.T, proxyID string) {
 	t.Helper()
@@ -115,7 +112,13 @@ func TestServer_StreamAggregatedResources_BasicProtocol(t *testing.T) {
 	envoy := NewTestEnvoy(t, "web-sidecar-proxy", "")
 	defer envoy.Close()
 
-	s := Server{logger, mgr, mgr, aclResolve}
+	s := Server{
+		Logger:       logger,
+		CfgMgr:       mgr,
+		Authz:        mgr,
+		ResolveToken: aclResolve,
+	}
+	s.Initialize()
 
 	go func() {
 		err := s.StreamAggregatedResources(envoy.stream)
@@ -162,11 +165,11 @@ func TestServer_StreamAggregatedResources_BasicProtocol(t *testing.T) {
 	// And should get a response immediately.
 	assertResponseSent(t, envoy.stream.sendCh, expectListenerJSON(t, snap, "", 1, 3))
 
-	// Now send Route request along with next listner one
+	// Now send Route request along with next listener one
 	envoy.SendReq(t, RouteType, 0, 0)
 	envoy.SendReq(t, ListenerType, 1, 3)
 
-	// We don't serve routes yet so this shoould block with no response
+	// We don't serve routes yet so this should block with no response
 	assertChanBlocked(t, envoy.stream.sendCh)
 
 	// WOOP! Envoy now has full connect config. Lets verify that if we update it,
@@ -175,11 +178,11 @@ func TestServer_StreamAggregatedResources_BasicProtocol(t *testing.T) {
 	// actually resend all blocked types on the new "version" anyway since it
 	// doesn't know _what_ changed. We could do something trivial but let's
 	// simulate a leaf cert expiring and being rotated.
-	snap.Leaf = proxycfg.TestLeafForCA(t, snap.Roots.Roots[0])
+	snap.ConnectProxy.Leaf = proxycfg.TestLeafForCA(t, snap.Roots.Roots[0])
 	mgr.DeliverConfig(t, "web-sidecar-proxy", snap)
 
 	// All 3 response that have something to return should return with new version
-	// note that the ordering is not determinisic in general. Trying to make this
+	// note that the ordering is not deterministic in general. Trying to make this
 	// test order-agnostic though is a massive pain since we are comparing
 	// non-identical JSON strings (so can simply sort by anything) and because we
 	// don't know the order the nonces will be assigned. For now we rely and
@@ -189,21 +192,21 @@ func TestServer_StreamAggregatedResources_BasicProtocol(t *testing.T) {
 	assertResponseSent(t, envoy.stream.sendCh, expectEndpointsJSON(t, snap, "", 2, 5))
 	assertResponseSent(t, envoy.stream.sendCh, expectListenerJSON(t, snap, "", 2, 6))
 
-	// Let's pretent that Envoy doesn't like that new listener config. It will ACK
+	// Let's pretend that Envoy doesn't like that new listener config. It will ACK
 	// all the others (same version) but NACK the listener. This is the most
 	// subtle part of xDS and the server implementation so I'll elaborate. A full
 	// description of the protocol can be found at
-	// https://github.com/envoyproxy/data-plane-api/blob/master/XDS_PROTOCOL.md.
-	// Envoy delays making a followup reqeest for a type until after it has
+	// https://www.envoyproxy.io/docs/envoy/latest/api-docs/xds_protocol
+	// Envoy delays making a followup request for a type until after it has
 	// processed and applied the last response. The next request then will include
-	// the nonce in the last response which acknowledges _recieving_ and handling
+	// the nonce in the last response which acknowledges _receiving_ and handling
 	// that response. It also includes the currently applied version. If all is
 	// good and it successfully applies the config, then the version in the next
 	// response will be the same version just sent. This is considered to be an
 	// ACK of that version for that type. If envoy fails to apply the config for
 	// some reason, it will still acknowledge that it received it (still return
 	// the responses nonce), but will show the previous version it's still using.
-	// This is considered a NACK. It's impotant that the server pay attention to
+	// This is considered a NACK. It's important that the server pay attention to
 	// the _nonce_ and not the version when deciding what to send otherwise a bad
 	// version that can't be applied in Envoy will cause a busy loop.
 	//
@@ -219,7 +222,7 @@ func TestServer_StreamAggregatedResources_BasicProtocol(t *testing.T) {
 	assertChanBlocked(t, envoy.stream.sendCh)
 
 	// Change config again and make sure it's delivered to everyone!
-	snap.Leaf = proxycfg.TestLeafForCA(t, snap.Roots.Roots[0])
+	snap.ConnectProxy.Leaf = proxycfg.TestLeafForCA(t, snap.Roots.Roots[0])
 	mgr.DeliverConfig(t, "web-sidecar-proxy", snap)
 
 	assertResponseSent(t, envoy.stream.sendCh, expectClustersJSON(t, snap, "", 3, 7))
@@ -227,140 +230,13 @@ func TestServer_StreamAggregatedResources_BasicProtocol(t *testing.T) {
 	assertResponseSent(t, envoy.stream.sendCh, expectListenerJSON(t, snap, "", 3, 9))
 }
 
-func expectListenerJSONResources(t *testing.T, snap *proxycfg.ConfigSnapshot, token string, v, n uint64) map[string]string {
-	tokenVal := ""
-	if token != "" {
-		tokenVal = fmt.Sprintf(",\n"+`"value": "%s"`, token)
-	}
-	return map[string]string{
-		"public_listener": `{
-													"@type": "type.googleapis.com/envoy.api.v2.Listener",
-													"name": "public_listener:0.0.0.0:9999",
-													"address": {
-														"socketAddress": {
-															"address": "0.0.0.0",
-															"portValue": 9999
-														}
-													},
-													"filterChains": [
-														{
-															"tlsContext": ` + expectedPublicTLSContextJSON(t, snap) + `,
-															"filters": [
-																{
-																	"name": "envoy.ext_authz",
-																	"config": {
-																			"grpc_service": {
-																					"envoy_grpc": {
-																						"cluster_name": "local_agent"
-																					},
-																					"initial_metadata": [
-																						{
-																							"key": "x-consul-token"
-																							` + tokenVal + `
-																						}
-																					]
-																				},
-																			"stat_prefix": "connect_authz"
-																		}
-																},
-																{
-																	"name": "envoy.tcp_proxy",
-																	"config": {
-																			"cluster": "local_app",
-																			"stat_prefix": "public_listener"
-																		}
-																}
-															]
-														}
-													]
-												}`,
-		"service:db": `{
-										"@type": "type.googleapis.com/envoy.api.v2.Listener",
-										"name": "service:db:127.0.0.1:9191",
-										"address": {
-											"socketAddress": {
-												"address": "127.0.0.1",
-												"portValue": 9191
-											}
-										},
-										"filterChains": [
-											{
-												"filters": [
-													{
-														"name": "envoy.tcp_proxy",
-														"config": {
-																"cluster": "service:db",
-																"stat_prefix": "service:db"
-															}
-													}
-												]
-											}
-										]
-									}`,
-		"prepared_query:geo-cache": `{
-																	"@type": "type.googleapis.com/envoy.api.v2.Listener",
-																	"name": "prepared_query:geo-cache:127.10.10.10:8181",
-																	"address": {
-																		"socketAddress": {
-																			"address": "127.10.10.10",
-																			"portValue": 8181
-																		}
-																	},
-																	"filterChains": [
-																		{
-																			"filters": [
-																				{
-																					"name": "envoy.tcp_proxy",
-																					"config": {
-																							"cluster": "prepared_query:geo-cache",
-																							"stat_prefix": "prepared_query:geo-cache"
-																						}
-																				}
-																			]
-																		}
-																	]
-																}`,
-	}
-}
-
-func expectListenerJSONFromResources(t *testing.T, snap *proxycfg.ConfigSnapshot, token string, v, n uint64, resourcesJSON map[string]string) string {
-	resJSON := ""
-	// Sort resources into specific order because that matters in JSONEq
-	// comparison later.
-	keyOrder := []string{"public_listener"}
-	for _, u := range snap.Proxy.Upstreams {
-		keyOrder = append(keyOrder, u.Identifier())
-	}
-	for _, k := range keyOrder {
-		j, ok := resourcesJSON[k]
-		if !ok {
-			continue
-		}
-		if resJSON != "" {
-			resJSON += ",\n"
-		}
-		resJSON += j
-	}
-	return `{
-		"versionInfo": "` + hexString(v) + `",
-		"resources": [` + resJSON + `],
-		"typeUrl": "type.googleapis.com/envoy.api.v2.Listener",
-		"nonce": "` + hexString(n) + `"
-		}`
-}
-
-func expectListenerJSON(t *testing.T, snap *proxycfg.ConfigSnapshot, token string, v, n uint64) string {
-	return expectListenerJSONFromResources(t, snap, token, v, n,
-		expectListenerJSONResources(t, snap, token, v, n))
-}
-
 func expectEndpointsJSON(t *testing.T, snap *proxycfg.ConfigSnapshot, token string, v, n uint64) string {
 	return `{
 		"versionInfo": "` + hexString(v) + `",
 		"resources": [
 			{
 				"@type": "type.googleapis.com/envoy.api.v2.ClusterLoadAssignment",
-				"clusterName": "service:db",
+				"clusterName": "db.default.dc1.internal.11111111-2222-3333-4444-555555555555.consul",
 				"endpoints": [
 					{
 						"lbEndpoints": [
@@ -369,20 +245,58 @@ func expectEndpointsJSON(t *testing.T, snap *proxycfg.ConfigSnapshot, token stri
 									"address": {
 										"socketAddress": {
 											"address": "10.10.1.1",
-											"portValue": 0
+											"portValue": 8080
 										}
 									}
-								}
+								},
+								"healthStatus": "HEALTHY",
+								"loadBalancingWeight": 1
 							},
 							{
 								"endpoint": {
 									"address": {
 										"socketAddress": {
 											"address": "10.10.1.2",
-											"portValue": 0
+											"portValue": 8080
 										}
 									}
-								}
+								},
+								"healthStatus": "HEALTHY",
+								"loadBalancingWeight": 1
+							}
+						]
+					}
+				]
+			},
+			{
+				"@type": "type.googleapis.com/envoy.api.v2.ClusterLoadAssignment",
+				"clusterName": "geo-cache.default.dc1.query.11111111-2222-3333-4444-555555555555.consul",
+				"endpoints": [
+					{
+						"lbEndpoints": [
+							{
+								"endpoint": {
+									"address": {
+										"socketAddress": {
+											"address": "10.10.1.1",
+											"portValue": 8080
+										}
+									}
+								},
+								"healthStatus": "HEALTHY",
+								"loadBalancingWeight": 1
+							},
+							{
+								"endpoint": {
+									"address": {
+										"socketAddress": {
+											"address": "10.10.1.2",
+											"portValue": 8080
+										}
+									}
+								},
+								"healthStatus": "HEALTHY",
+								"loadBalancingWeight": 1
 							}
 						]
 					}
@@ -394,67 +308,15 @@ func expectEndpointsJSON(t *testing.T, snap *proxycfg.ConfigSnapshot, token stri
 	}`
 }
 
-func expectClustersJSON(t *testing.T, snap *proxycfg.ConfigSnapshot, token string, v, n uint64) string {
-	return `{
-		"versionInfo": "` + hexString(v) + `",
-		"resources": [
-			{
-				"@type": "type.googleapis.com/envoy.api.v2.Cluster",
-				"name": "local_app",
-				"connectTimeout": "5s",
-				"hosts": [
-					{
-						"socketAddress": {
-							"address": "127.0.0.1",
-							"portValue": 8080
-						}
-					}
-				]
-			},
-			{
-				"@type": "type.googleapis.com/envoy.api.v2.Cluster",
-				"name": "service:db",
-				"type": "EDS",
-				"edsClusterConfig": {
-					"edsConfig": {
-						"ads": {
-
-						}
-					}
-				},
-				"connectTimeout": "5s",
-				"tlsContext": ` + expectedUpstreamTLSContextJSON(t, snap) + `
-			},
-			{
-				"@type": "type.googleapis.com/envoy.api.v2.Cluster",
-				"name": "prepared_query:geo-cache",
-				"type": "EDS",
-				"edsClusterConfig": {
-					"edsConfig": {
-						"ads": {
-
-						}
-					}
-				},
-				"connectTimeout": "5s",
-				"tlsContext": ` + expectedUpstreamTLSContextJSON(t, snap) + `
-			}
-		],
-		"typeUrl": "type.googleapis.com/envoy.api.v2.Cluster",
-		"nonce": "` + hexString(n) + `"
-	}
-	`
-}
-
-func expectedUpstreamTLSContextJSON(t *testing.T, snap *proxycfg.ConfigSnapshot) string {
-	return expectedTLSContextJSON(t, snap, false)
+func expectedUpstreamTLSContextJSON(t *testing.T, snap *proxycfg.ConfigSnapshot, sni string) string {
+	return expectedTLSContextJSON(t, snap, false, sni)
 }
 
 func expectedPublicTLSContextJSON(t *testing.T, snap *proxycfg.ConfigSnapshot) string {
-	return expectedTLSContextJSON(t, snap, true)
+	return expectedTLSContextJSON(t, snap, true, "")
 }
 
-func expectedTLSContextJSON(t *testing.T, snap *proxycfg.ConfigSnapshot, requireClientCert bool) string {
+func expectedTLSContextJSON(t *testing.T, snap *proxycfg.ConfigSnapshot, requireClientCert bool, sni string) string {
 	// Assume just one root for now, can get fancier later if needed.
 	caPEM := snap.Roots.Roots[0].RootCert
 	reqClient := ""
@@ -462,16 +324,23 @@ func expectedTLSContextJSON(t *testing.T, snap *proxycfg.ConfigSnapshot, require
 		reqClient = `,
 		"requireClientCertificate": true`
 	}
+
+	upstreamSNI := ""
+	if sni != "" {
+		upstreamSNI = `,
+		"sni": "` + sni + `"`
+	}
+
 	return `{
 		"commonTlsContext": {
 			"tlsParams": {},
 			"tlsCertificates": [
 				{
 					"certificateChain": {
-						"inlineString": "` + strings.Replace(snap.Leaf.CertPEM, "\n", "\\n", -1) + `"
+						"inlineString": "` + strings.Replace(snap.ConnectProxy.Leaf.CertPEM, "\n", "\\n", -1) + `"
 					},
 					"privateKey": {
-						"inlineString": "` + strings.Replace(snap.Leaf.PrivateKeyPEM, "\n", "\\n", -1) + `"
+						"inlineString": "` + strings.Replace(snap.ConnectProxy.Leaf.PrivateKeyPEM, "\n", "\\n", -1) + `"
 					}
 				}
 			],
@@ -482,6 +351,7 @@ func expectedTLSContextJSON(t *testing.T, snap *proxycfg.ConfigSnapshot, require
 			}
 		}
 		` + reqClient + `
+		` + upstreamSNI + `
 	}`
 }
 
@@ -489,7 +359,7 @@ func assertChanBlocked(t *testing.T, ch chan *envoy.DiscoveryResponse) {
 	t.Helper()
 	select {
 	case r := <-ch:
-		t.Fatalf("chan should block but recieved: %v", r)
+		t.Fatalf("chan should block but received: %v", r)
 	case <-time.After(10 * time.Millisecond):
 		return
 	}
@@ -501,25 +371,21 @@ func assertResponseSent(t *testing.T, ch chan *envoy.DiscoveryResponse, wantJSON
 	case r := <-ch:
 		assertResponse(t, r, wantJSON)
 	case <-time.After(50 * time.Millisecond):
-		t.Fatalf("no response recieved after 50ms")
+		t.Fatalf("no response received after 50ms")
 	}
 }
 
 // assertResponse is a helper to test a envoy.DiscoveryResponse matches the
-// JSON representaion we expect. We use JSON because the responses use protobuf
+// JSON representation we expect. We use JSON because the responses use protobuf
 // Any type which includes binary protobuf encoding and would make creating
 // expected structs require the same code that is under test!
 func assertResponse(t *testing.T, r *envoy.DiscoveryResponse, wantJSON string) {
 	t.Helper()
-	m := jsonpb.Marshaler{
-		Indent: "  ",
-	}
-	gotJSON, err := m.MarshalToString(r)
-	require.NoError(t, err)
+	gotJSON := responseToJSON(t, r)
 	require.JSONEqf(t, wantJSON, gotJSON, "got:\n%s", gotJSON)
 }
 
-func TestServer_StreamAggregatedResources_ACLEnforcment(t *testing.T) {
+func TestServer_StreamAggregatedResources_ACLEnforcement(t *testing.T) {
 
 	tests := []struct {
 		name        string
@@ -529,7 +395,7 @@ func TestServer_StreamAggregatedResources_ACLEnforcment(t *testing.T) {
 		wantDenied  bool
 	}{
 		// Note that although we've stubbed actual ACL checks in the testManager
-		// ConnectAuthorize mock, by asserting against specifc reason strings here
+		// ConnectAuthorize mock, by asserting against specific reason strings here
 		// even in the happy case which can't match the default one returned by the
 		// mock we are implicitly validating that the implementation used the
 		// correct token from the context.
@@ -544,21 +410,21 @@ func TestServer_StreamAggregatedResources_ACLEnforcment(t *testing.T) {
 			wantDenied:  true,
 		},
 		{
-			name:        "default deny, service:write token",
+			name:        "default deny, write token",
 			defaultDeny: true,
 			acl:         `service "web" { policy = "write" }`,
 			token:       "service-write-on-web",
 			wantDenied:  false,
 		},
 		{
-			name:        "default deny, service:read token",
+			name:        "default deny, read token",
 			defaultDeny: true,
 			acl:         `service "web" { policy = "read" }`,
 			token:       "service-write-on-web",
 			wantDenied:  true,
 		},
 		{
-			name:        "default deny, service:write token on different service",
+			name:        "default deny, write token on different service",
 			defaultDeny: true,
 			acl:         `service "not-web" { policy = "write" }`,
 			token:       "service-write-on-not-web",
@@ -589,7 +455,13 @@ func TestServer_StreamAggregatedResources_ACLEnforcment(t *testing.T) {
 			envoy := NewTestEnvoy(t, "web-sidecar-proxy", tt.token)
 			defer envoy.Close()
 
-			s := Server{logger, mgr, mgr, aclResolve}
+			s := Server{
+				Logger:       logger,
+				CfgMgr:       mgr,
+				Authz:        mgr,
+				ResolveToken: aclResolve,
+			}
+			s.Initialize()
 
 			errCh := make(chan error, 1)
 			go func() {
@@ -632,6 +504,196 @@ func TestServer_StreamAggregatedResources_ACLEnforcment(t *testing.T) {
 	}
 }
 
+func TestServer_StreamAggregatedResources_ACLTokenDeleted_StreamTerminatedDuringDiscoveryRequest(t *testing.T) {
+	aclRules := `service "web" { policy = "write" }`
+	token := "service-write-on-web"
+
+	policy, err := acl.NewPolicyFromSource("", 0, aclRules, acl.SyntaxLegacy, nil)
+	require.NoError(t, err)
+
+	var validToken atomic.Value
+	validToken.Store(token)
+
+	logger := log.New(os.Stderr, "", log.LstdFlags)
+	mgr := newTestManager(t)
+	aclResolve := func(id string) (acl.Authorizer, error) {
+		if token := validToken.Load(); token == nil || id != token.(string) {
+			return nil, acl.ErrNotFound
+		}
+
+		return acl.NewPolicyAuthorizer(acl.RootAuthorizer("deny"), []*acl.Policy{policy}, nil)
+	}
+	envoy := NewTestEnvoy(t, "web-sidecar-proxy", token)
+	defer envoy.Close()
+
+	s := Server{
+		Logger:             logger,
+		CfgMgr:             mgr,
+		Authz:              mgr,
+		ResolveToken:       aclResolve,
+		AuthCheckFrequency: 1 * time.Hour, // make sure this doesn't kick in
+	}
+	s.Initialize()
+
+	errCh := make(chan error, 1)
+	go func() {
+		errCh <- s.StreamAggregatedResources(envoy.stream)
+	}()
+
+	getError := func() (gotErr error, ok bool) {
+		select {
+		case err := <-errCh:
+			return err, true
+		default:
+			return nil, false
+		}
+	}
+
+	// Register the proxy to create state needed to Watch() on
+	mgr.RegisterProxy(t, "web-sidecar-proxy")
+
+	// Send initial cluster discover (OK)
+	envoy.SendReq(t, ClusterType, 0, 0)
+	{
+		err, ok := getError()
+		require.NoError(t, err)
+		require.False(t, ok)
+	}
+
+	// Check no response sent yet
+	assertChanBlocked(t, envoy.stream.sendCh)
+	{
+		err, ok := getError()
+		require.NoError(t, err)
+		require.False(t, ok)
+	}
+
+	// Deliver a new snapshot
+	snap := proxycfg.TestConfigSnapshot(t)
+	mgr.DeliverConfig(t, "web-sidecar-proxy", snap)
+
+	assertResponseSent(t, envoy.stream.sendCh, expectClustersJSON(t, snap, token, 1, 1))
+
+	// Now nuke the ACL token.
+	validToken.Store("")
+
+	// It also (in parallel) issues the next cluster request (which acts as an ACK
+	// of the version we sent)
+	envoy.SendReq(t, ClusterType, 1, 1)
+
+	select {
+	case err := <-errCh:
+		require.Error(t, err)
+		gerr, ok := status.FromError(err)
+		require.Truef(t, ok, "not a grpc status error: type='%T' value=%v", err, err)
+		require.Equal(t, codes.Unauthenticated, gerr.Code())
+		require.Equal(t, "unauthenticated: ACL not found", gerr.Message())
+
+		mgr.AssertWatchCancelled(t, "web-sidecar-proxy")
+	case <-time.After(50 * time.Millisecond):
+		t.Fatalf("timed out waiting for handler to finish")
+	}
+}
+
+func TestServer_StreamAggregatedResources_ACLTokenDeleted_StreamTerminatedInBackground(t *testing.T) {
+	aclRules := `service "web" { policy = "write" }`
+	token := "service-write-on-web"
+
+	policy, err := acl.NewPolicyFromSource("", 0, aclRules, acl.SyntaxLegacy, nil)
+	require.NoError(t, err)
+
+	var validToken atomic.Value
+	validToken.Store(token)
+
+	logger := log.New(os.Stderr, "", log.LstdFlags)
+	mgr := newTestManager(t)
+	aclResolve := func(id string) (acl.Authorizer, error) {
+		if token := validToken.Load(); token == nil || id != token.(string) {
+			return nil, acl.ErrNotFound
+		}
+
+		return acl.NewPolicyAuthorizer(acl.RootAuthorizer("deny"), []*acl.Policy{policy}, nil)
+	}
+	envoy := NewTestEnvoy(t, "web-sidecar-proxy", token)
+	defer envoy.Close()
+
+	s := Server{
+		Logger:             logger,
+		CfgMgr:             mgr,
+		Authz:              mgr,
+		ResolveToken:       aclResolve,
+		AuthCheckFrequency: 100 * time.Millisecond, // Make this short.
+	}
+	s.Initialize()
+
+	errCh := make(chan error, 1)
+	go func() {
+		errCh <- s.StreamAggregatedResources(envoy.stream)
+	}()
+
+	getError := func() (gotErr error, ok bool) {
+		select {
+		case err := <-errCh:
+			return err, true
+		default:
+			return nil, false
+		}
+	}
+
+	// Register the proxy to create state needed to Watch() on
+	mgr.RegisterProxy(t, "web-sidecar-proxy")
+
+	// Send initial cluster discover (OK)
+	envoy.SendReq(t, ClusterType, 0, 0)
+	{
+		err, ok := getError()
+		require.NoError(t, err)
+		require.False(t, ok)
+	}
+
+	// Check no response sent yet
+	assertChanBlocked(t, envoy.stream.sendCh)
+	{
+		err, ok := getError()
+		require.NoError(t, err)
+		require.False(t, ok)
+	}
+
+	// Deliver a new snapshot
+	snap := proxycfg.TestConfigSnapshot(t)
+	mgr.DeliverConfig(t, "web-sidecar-proxy", snap)
+
+	assertResponseSent(t, envoy.stream.sendCh, expectClustersJSON(t, snap, token, 1, 1))
+
+	// It also (in parallel) issues the next cluster request (which acts as an ACK
+	// of the version we sent)
+	envoy.SendReq(t, ClusterType, 1, 1)
+
+	// Check no response sent yet
+	assertChanBlocked(t, envoy.stream.sendCh)
+	{
+		err, ok := getError()
+		require.NoError(t, err)
+		require.False(t, ok)
+	}
+
+	// Now nuke the ACL token while there's no activity.
+	validToken.Store("")
+
+	select {
+	case err := <-errCh:
+		require.Error(t, err)
+		gerr, ok := status.FromError(err)
+		require.Truef(t, ok, "not a grpc status error: type='%T' value=%v", err, err)
+		require.Equal(t, codes.Unauthenticated, gerr.Code())
+		require.Equal(t, "unauthenticated: ACL not found", gerr.Message())
+
+		mgr.AssertWatchCancelled(t, "web-sidecar-proxy")
+	case <-time.After(200 * time.Millisecond):
+		t.Fatalf("timed out waiting for handler to finish")
+	}
+}
+
 // This tests the ext_authz service method that implements connect authz.
 func TestServer_Check(t *testing.T) {
 
@@ -685,7 +747,7 @@ func TestServer_Check(t *testing.T) {
 			destPrincipal: "not-a-spiffe-id",
 			// Should never make it to authz call.
 			wantDenied: true,
-			wantReason: "Destination Principal is not a valid Connect identitiy",
+			wantReason: "Destination Principal is not a valid Connect identity",
 		},
 		{
 			name:          "dest not a service URI",
@@ -693,7 +755,7 @@ func TestServer_Check(t *testing.T) {
 			destPrincipal: "spiffe://trust-domain.consul",
 			// Should never make it to authz call.
 			wantDenied: true,
-			wantReason: "Destination Principal is not a valid Service identitiy",
+			wantReason: "Destination Principal is not a valid Service identity",
 		},
 		{
 			name:        "ACL not got permission for authz call",
@@ -729,7 +791,13 @@ func TestServer_Check(t *testing.T) {
 			envoy := NewTestEnvoy(t, "web-sidecar-proxy", token)
 			defer envoy.Close()
 
-			s := Server{logger, mgr, mgr, aclResolve}
+			s := Server{
+				Logger:       logger,
+				CfgMgr:       mgr,
+				Authz:        mgr,
+				ResolveToken: aclResolve,
+			}
+			s.Initialize()
 
 			// Create a context with the correct token
 			ctx := metadata.NewIncomingContext(context.Background(),
@@ -763,213 +831,3 @@ func TestServer_Check(t *testing.T) {
 		})
 	}
 }
-
-func TestServer_ConfigOverrides(t *testing.T) {
-
-	tests := []struct {
-		name  string
-		setup func(snap *proxycfg.ConfigSnapshot) string
-	}{
-		{
-			name: "sanity check no custom",
-			setup: func(snap *proxycfg.ConfigSnapshot) string {
-				// Default snap and expectation
-				return expectListenerJSON(t, snap, "my-token", 1, 1)
-			},
-		},
-		{
-			name: "custom public_listener no type",
-			setup: func(snap *proxycfg.ConfigSnapshot) string {
-				snap.Proxy.Config["envoy_public_listener_json"] =
-					customListenerJSON(t, customListenerJSONOptions{
-						Name:        "custom-public-listen",
-						IncludeType: false,
-					})
-				resources := expectListenerJSONResources(t, snap, "my-token", 1, 1)
-
-				// Replace the public listener with the custom one WITH type since
-				// that's how it comes out the other end, and with TLS and authz
-				// overridden.
-				resources["public_listener"] = customListenerJSON(t, customListenerJSONOptions{
-					Name: "custom-public-listen",
-					// We should add type, TLS and authz
-					IncludeType:   true,
-					OverrideAuthz: true,
-					TLSContext:    expectedPublicTLSContextJSON(t, snap),
-				})
-				return expectListenerJSONFromResources(t, snap, "my-token", 1, 1, resources)
-			},
-		},
-		{
-			name: "custom public_listener with type",
-			setup: func(snap *proxycfg.ConfigSnapshot) string {
-				snap.Proxy.Config["envoy_public_listener_json"] =
-					customListenerJSON(t, customListenerJSONOptions{
-						Name:        "custom-public-listen",
-						IncludeType: true,
-					})
-				resources := expectListenerJSONResources(t, snap, "my-token", 1, 1)
-
-				// Replace the public listener with the custom one WITH type since
-				// that's how it comes out the other end, and with TLS and authz
-				// overridden.
-				resources["public_listener"] = customListenerJSON(t, customListenerJSONOptions{
-					Name: "custom-public-listen",
-					// We should add type, TLS and authz
-					IncludeType:   true,
-					OverrideAuthz: true,
-					TLSContext:    expectedPublicTLSContextJSON(t, snap),
-				})
-				return expectListenerJSONFromResources(t, snap, "my-token", 1, 1, resources)
-			},
-		},
-		{
-			name: "custom public_listener with TLS should be overridden",
-			setup: func(snap *proxycfg.ConfigSnapshot) string {
-				snap.Proxy.Config["envoy_public_listener_json"] =
-					customListenerJSON(t, customListenerJSONOptions{
-						Name:        "custom-public-listen",
-						IncludeType: true,
-						TLSContext:  `{"requireClientCertificate": false}`,
-					})
-				resources := expectListenerJSONResources(t, snap, "my-token", 1, 1)
-
-				// Replace the public listener with the custom one WITH type since
-				// that's how it comes out the other end, and with TLS and authz
-				// overridden.
-				resources["public_listener"] = customListenerJSON(t, customListenerJSONOptions{
-					Name: "custom-public-listen",
-					// We should add type, TLS and authz
-					IncludeType:   true,
-					OverrideAuthz: true,
-					TLSContext:    expectedPublicTLSContextJSON(t, snap),
-				})
-				return expectListenerJSONFromResources(t, snap, "my-token", 1, 1, resources)
-			},
-		},
-		{
-			name: "custom upstream no type",
-			setup: func(snap *proxycfg.ConfigSnapshot) string {
-				snap.Proxy.Upstreams[0].Config["envoy_listener_json"] =
-					customListenerJSON(t, customListenerJSONOptions{
-						Name:        "custom-upstream",
-						IncludeType: false,
-					})
-				resources := expectListenerJSONResources(t, snap, "my-token", 1, 1)
-
-				// Replace an upstream listener with the custom one WITH type since
-				// that's how it comes out the other end. Note we do override TLS
-				resources["service:db"] =
-					customListenerJSON(t, customListenerJSONOptions{
-						Name: "custom-upstream",
-						// We should add type
-						IncludeType: true,
-					})
-				return expectListenerJSONFromResources(t, snap, "my-token", 1, 1, resources)
-			},
-		},
-		{
-			name: "custom upstream with type",
-			setup: func(snap *proxycfg.ConfigSnapshot) string {
-				snap.Proxy.Upstreams[0].Config["envoy_listener_json"] =
-					customListenerJSON(t, customListenerJSONOptions{
-						Name:        "custom-upstream",
-						IncludeType: true,
-					})
-				resources := expectListenerJSONResources(t, snap, "my-token", 1, 1)
-
-				// Replace an upstream listener with the custom one WITH type since
-				// that's how it comes out the other end.
-				resources["service:db"] =
-					customListenerJSON(t, customListenerJSONOptions{
-						Name: "custom-upstream",
-						// We should add type
-						IncludeType: true,
-					})
-				return expectListenerJSONFromResources(t, snap, "my-token", 1, 1, resources)
-			},
-		},
-	}
-
-	for _, tt := range tests {
-		t.Run(tt.name, func(t *testing.T) {
-			require := require.New(t)
-
-			// Sanity check default with no overrides first
-			snap := proxycfg.TestConfigSnapshot(t)
-			expect := tt.setup(snap)
-
-			listeners, err := listenersFromSnapshot(snap, "my-token")
-			require.NoError(err)
-			r, err := createResponse(ListenerType, "00000001", "00000001", listeners)
-			require.NoError(err)
-
-			assertResponse(t, r, expect)
-		})
-	}
-}
-
-type customListenerJSONOptions struct {
-	Name          string
-	IncludeType   bool
-	OverrideAuthz bool
-	TLSContext    string
-}
-
-const customListenerJSONTpl = `{
-	{{ if .IncludeType -}}
-	"@type": "type.googleapis.com/envoy.api.v2.Listener",
-	{{- end }}
-	"name": "{{ .Name }}",
-	"address": {
-		"socketAddress": {
-			"address": "11.11.11.11",
-			"portValue": 11111
-		}
-	},
-	"filterChains": [
-		{
-			{{ if .TLSContext -}}
-			"tlsContext": {{ .TLSContext }},
-			{{- end }}
-			"filters": [
-				{{ if .OverrideAuthz -}}
-				{
-					"name": "envoy.ext_authz",
-					"config": {
-							"grpc_service": {
-										"envoy_grpc": {
-													"cluster_name": "local_agent"
-												},
-										"initial_metadata": [
-													{
-																"key": "x-consul-token",
-																"value": "my-token"
-															}
-												]
-									},
-							"stat_prefix": "connect_authz"
-						}
-				},
-				{{- end }}
-				{
-					"name": "envoy.tcp_proxy",
-					"config": {
-							"cluster": "random-cluster",
-							"stat_prefix": "foo-stats"
-						}
-				}
-			]
-		}
-	]
-}`
-
-var customListenerJSONTemplate = template.Must(template.New("").Parse(customListenerJSONTpl))
-
-func customListenerJSON(t *testing.T, opts customListenerJSONOptions) string {
-	t.Helper()
-	var buf bytes.Buffer
-	err := customListenerJSONTemplate.Execute(&buf, opts)
-	require.NoError(t, err)
-	return buf.String()
-}
