@@ -1,4 +1,3 @@
-
 /**
  *    Copyright (C) 2018-present MongoDB, Inc.
  *
@@ -28,26 +27,28 @@
  *    it in the license file.
  */
 
-#define MONGO_LOG_DEFAULT_COMPONENT ::mongo::logger::LogComponent::kSharding
+#define MONGO_LOGV2_DEFAULT_COMPONENT ::mongo::logv2::LogComponent::kSharding
 
 #include "mongo/platform/basic.h"
 
 #include "mongo/db/s/balancer/balancer_chunk_selection_policy_impl.h"
 
 #include <algorithm>
+#include <memory>
 #include <vector>
 
 #include "mongo/base/status_with.h"
 #include "mongo/bson/bsonobj_comparator_interface.h"
-#include "mongo/s/catalog/sharding_catalog_client.h"
+#include "mongo/db/s/sharding_runtime_d_params_gen.h"
+#include "mongo/logv2/log.h"
+#include "mongo/platform/bits.h"
+#include "mongo/s/balancer_configuration.h"
 #include "mongo/s/catalog/type_chunk.h"
 #include "mongo/s/catalog/type_collection.h"
 #include "mongo/s/catalog/type_tags.h"
 #include "mongo/s/catalog_cache.h"
 #include "mongo/s/grid.h"
-#include "mongo/stdx/memory.h"
-#include "mongo/util/log.h"
-#include "mongo/util/mongoutils/str.h"
+#include "mongo/util/str.h"
 
 namespace mongo {
 
@@ -76,27 +77,24 @@ StatusWith<DistributionStatus> createCollectionDistributionStatus(
     for (const auto& chunkEntry : chunkMgr->chunks()) {
         ChunkType chunk;
         chunk.setNS(chunkMgr->getns());
-        chunk.setMin(chunkEntry->getMin());
-        chunk.setMax(chunkEntry->getMax());
-        chunk.setJumbo(chunkEntry->isJumbo());
-        chunk.setShard(chunkEntry->getShardId());
-        chunk.setVersion(chunkEntry->getLastmod());
+        chunk.setMin(chunkEntry.getMin());
+        chunk.setMax(chunkEntry.getMax());
+        chunk.setJumbo(chunkEntry.isJumbo());
+        chunk.setShard(chunkEntry.getShardId());
+        chunk.setVersion(chunkEntry.getLastmod());
 
-        shardToChunksMap[chunkEntry->getShardId()].push_back(chunk);
+        shardToChunksMap[chunkEntry.getShardId()].push_back(chunk);
     }
 
-    vector<TagsType> collectionTags;
-    Status tagsStatus = Grid::get(opCtx)->catalogClient()->getTagsForCollection(
-        opCtx, chunkMgr->getns(), &collectionTags);
-    if (!tagsStatus.isOK()) {
-        return {tagsStatus.code(),
-                str::stream() << "Unable to load tags for collection " << chunkMgr->getns()
-                              << " due to "
-                              << tagsStatus.toString()};
+    const auto swCollectionTags =
+        Grid::get(opCtx)->catalogClient()->getTagsForCollection(opCtx, chunkMgr->getns());
+    if (!swCollectionTags.isOK()) {
+        return swCollectionTags.getStatus().withContext(
+            str::stream() << "Unable to load tags for collection " << chunkMgr->getns().ns());
     }
+    const auto& collectionTags = swCollectionTags.getValue();
 
-    DistributionStatus distribution(NamespaceString(chunkMgr->getns()),
-                                    std::move(shardToChunksMap));
+    DistributionStatus distribution(chunkMgr->getns(), std::move(shardToChunksMap));
 
     // Cache the collection tags
     const auto& keyPattern = chunkMgr->getShardKeyPattern().getKeyPattern();
@@ -122,7 +120,8 @@ StatusWith<DistributionStatus> createCollectionDistributionStatus(
  * splitting a chunk does not yield the same chunk anymore.
  */
 class SplitCandidatesBuffer {
-    MONGO_DISALLOW_COPYING(SplitCandidatesBuffer);
+    SplitCandidatesBuffer(const SplitCandidatesBuffer&) = delete;
+    SplitCandidatesBuffer& operator=(const SplitCandidatesBuffer&) = delete;
 
 public:
     SplitCandidatesBuffer(NamespaceString nss, ChunkVersion collectionVersion)
@@ -136,16 +135,16 @@ public:
      * Adds the specified split point to the chunk. The split points must always be within the
      * boundaries of the chunk and must come in increasing order.
      */
-    void addSplitPoint(shared_ptr<Chunk> chunk, const BSONObj& splitPoint) {
-        auto it = _chunkSplitPoints.find(chunk->getMin());
+    void addSplitPoint(const Chunk& chunk, const BSONObj& splitPoint) {
+        auto it = _chunkSplitPoints.find(chunk.getMin());
         if (it == _chunkSplitPoints.end()) {
-            _chunkSplitPoints.emplace(chunk->getMin(),
-                                      BalancerChunkSelectionPolicy::SplitInfo(chunk->getShardId(),
+            _chunkSplitPoints.emplace(chunk.getMin(),
+                                      BalancerChunkSelectionPolicy::SplitInfo(chunk.getShardId(),
                                                                               _nss,
                                                                               _collectionVersion,
-                                                                              chunk->getLastmod(),
-                                                                              chunk->getMin(),
-                                                                              chunk->getMax(),
+                                                                              chunk.getLastmod(),
+                                                                              chunk.getMin(),
+                                                                              chunk.getMax(),
                                                                               {splitPoint}));
         } else if (splitPoint.woCompare(it->second.splitKeys.back()) > 0) {
             it->second.splitKeys.push_back(splitPoint);
@@ -177,6 +176,95 @@ private:
     BSONObjIndexedMap<BalancerChunkSelectionPolicy::SplitInfo> _chunkSplitPoints;
 };
 
+/**
+ * Populates splitCandidates with chunk and splitPoint pairs for chunks that violate tag
+ * range boundaries.
+ */
+void getSplitCandidatesToEnforceTagRanges(const ChunkManager* cm,
+                                          const DistributionStatus& distribution,
+                                          SplitCandidatesBuffer* splitCandidates) {
+    const auto& globalMax = cm->getShardKeyPattern().getKeyPattern().globalMax();
+
+    // For each tag range, find chunks that need to be split.
+    for (const auto& tagRangeEntry : distribution.tagRanges()) {
+        const auto& tagRange = tagRangeEntry.second;
+
+        const auto chunkAtZoneMin = cm->findIntersectingChunkWithSimpleCollation(tagRange.min);
+        invariant(chunkAtZoneMin.getMax().woCompare(tagRange.min) > 0);
+
+        if (chunkAtZoneMin.getMin().woCompare(tagRange.min)) {
+            splitCandidates->addSplitPoint(chunkAtZoneMin, tagRange.min);
+        }
+
+        // The global max key can never fall in the middle of a chunk.
+        if (!tagRange.max.woCompare(globalMax))
+            continue;
+
+        const auto chunkAtZoneMax = cm->findIntersectingChunkWithSimpleCollation(tagRange.max);
+
+        // We need to check that both the chunk's minKey does not match the zone's max and also that
+        // the max is not equal, which would only happen in the case of the zone ending in MaxKey.
+        if (chunkAtZoneMax.getMin().woCompare(tagRange.max) &&
+            chunkAtZoneMax.getMax().woCompare(tagRange.max)) {
+            splitCandidates->addSplitPoint(chunkAtZoneMax, tagRange.max);
+        }
+    }
+}
+
+/**
+ * If the number of chunks as given by the ChunkManager is less than the configured minimum
+ * number of chunks for the sessions collection (minNumChunksForSessionsCollection), calculates
+ * split points that evenly partition the key space into N ranges (where N is
+ * minNumChunksForSessionsCollection rounded up to the next power of 2), and populates
+ * splitCandidates with chunk and splitPoint pairs for chunks that need to split.
+ */
+void getSplitCandidatesForSessionsCollection(OperationContext* opCtx,
+                                             const ChunkManager* cm,
+                                             SplitCandidatesBuffer* splitCandidates) {
+    const auto minNumChunks = minNumChunksForSessionsCollection.load();
+
+    if (cm->numChunks() >= minNumChunks) {
+        return;
+    }
+
+    // Use the next power of 2 as the target number of chunks.
+    const size_t numBits = 64 - countLeadingZeros64(minNumChunks - 1);
+    const uint32_t numChunks = 1 << numBits;
+
+    // Compute split points for _id.id that partition the UUID 128-bit data space into numChunks
+    // equal ranges. Since the numChunks is a power of 2, the split points are the permutations
+    // of the prefix numBits right-padded with 0's.
+    std::vector<BSONObj> splitPoints;
+    for (uint32_t i = 1; i < numChunks; i++) {
+        // Start with a buffer of 0's.
+        std::array<uint8_t, 16> buf{0b0};
+
+        // Left-shift i to fill the remaining bits in the prefix 32 bits with 0's.
+        const uint32_t high = i << (CHAR_BIT * 4 - numBits);
+
+        // Fill the prefix 4 bytes with high's bytes.
+        buf[0] = static_cast<uint8_t>(high >> CHAR_BIT * 3);
+        buf[1] = static_cast<uint8_t>(high >> CHAR_BIT * 2);
+        buf[2] = static_cast<uint8_t>(high >> CHAR_BIT * 1);
+        buf[3] = static_cast<uint8_t>(high);
+
+        ConstDataRange cdr(buf.data(), sizeof(buf));
+        splitPoints.push_back(BSON("_id" << BSON("id" << UUID::fromCDR(cdr))));
+    }
+
+    // For each split point, find a chunk that needs to be split.
+    for (auto& splitPoint : splitPoints) {
+        const auto chunkAtSplitPoint = cm->findIntersectingChunkWithSimpleCollation(splitPoint);
+        invariant(chunkAtSplitPoint.getMax().woCompare(splitPoint) > 0);
+
+        if (chunkAtSplitPoint.getMin().woCompare(splitPoint)) {
+            splitCandidates->addSplitPoint(chunkAtSplitPoint, splitPoint);
+        }
+    }
+
+    return;
+}
+
 }  // namespace
 
 BalancerChunkSelectionPolicyImpl::BalancerChunkSelectionPolicyImpl(ClusterStatistics* clusterStats,
@@ -192,16 +280,15 @@ StatusWith<SplitInfoVector> BalancerChunkSelectionPolicyImpl::selectChunksToSpli
         return shardStatsStatus.getStatus();
     }
 
-    const auto shardStats = std::move(shardStatsStatus.getValue());
+    const auto& shardStats = shardStatsStatus.getValue();
 
-    vector<CollectionType> collections;
-
-    Status collsStatus =
-        Grid::get(opCtx)->catalogClient()->getCollections(opCtx, nullptr, &collections, nullptr);
-    if (!collsStatus.isOK()) {
-        return collsStatus;
+    auto swCollections = Grid::get(opCtx)->catalogClient()->getCollections(opCtx, nullptr, nullptr);
+    if (!swCollections.isOK()) {
+        return swCollections.getStatus();
     }
 
+    auto& collections = swCollections.getValue();
+
     if (collections.empty()) {
         return SplitInfoVector{};
     }
@@ -222,8 +309,20 @@ StatusWith<SplitInfoVector> BalancerChunkSelectionPolicyImpl::selectChunksToSpli
             // Namespace got dropped before we managed to get to it, so just skip it
             continue;
         } else if (!candidatesStatus.isOK()) {
-            warning() << "Unable to enforce tag range policy for collection " << nss.ns()
-                      << causedBy(candidatesStatus.getStatus());
+            if (nss == NamespaceString::kLogicalSessionsNamespace) {
+                LOGV2_WARNING(4562402,
+                              "Unable to split sessions collection chunks",
+                              "error"_attr = candidatesStatus.getStatus());
+
+            } else {
+                LOGV2_WARNING(
+                    21852,
+                    "Unable to enforce tag range policy for collection {namespace}: {error}",
+                    "Unable to enforce tag range policy for collection",
+                    "namespace"_attr = nss.ns(),
+                    "error"_attr = candidatesStatus.getStatus());
+            }
+
             continue;
         }
 
@@ -235,27 +334,39 @@ StatusWith<SplitInfoVector> BalancerChunkSelectionPolicyImpl::selectChunksToSpli
     return splitCandidates;
 }
 
-StatusWith<MigrateInfoVector> BalancerChunkSelectionPolicyImpl::selectChunksToMove(
-    OperationContext* opCtx, bool aggressiveBalanceHint) {
+StatusWith<SplitInfoVector> BalancerChunkSelectionPolicyImpl::selectChunksToSplit(
+    OperationContext* opCtx, const NamespaceString& nss) {
+
     auto shardStatsStatus = _clusterStats->getStats(opCtx);
     if (!shardStatsStatus.isOK()) {
         return shardStatsStatus.getStatus();
     }
 
-    const auto shardStats = std::move(shardStatsStatus.getValue());
+    const auto& shardStats = shardStatsStatus.getValue();
+
+    return _getSplitCandidatesForCollection(opCtx, nss, shardStats);
+}
+
+StatusWith<MigrateInfoVector> BalancerChunkSelectionPolicyImpl::selectChunksToMove(
+    OperationContext* opCtx) {
+    auto shardStatsStatus = _clusterStats->getStats(opCtx);
+    if (!shardStatsStatus.isOK()) {
+        return shardStatsStatus.getStatus();
+    }
+
+    const auto& shardStats = shardStatsStatus.getValue();
 
     if (shardStats.size() < 2) {
         return MigrateInfoVector{};
     }
 
-    vector<CollectionType> collections;
-
-    Status collsStatus =
-        Grid::get(opCtx)->catalogClient()->getCollections(opCtx, nullptr, &collections, nullptr);
-    if (!collsStatus.isOK()) {
-        return collsStatus;
+    auto swCollections = Grid::get(opCtx)->catalogClient()->getCollections(opCtx, nullptr, nullptr);
+    if (!swCollections.isOK()) {
+        return swCollections.getStatus();
     }
 
+    auto& collections = swCollections.getValue();
+
     if (collections.empty()) {
         return MigrateInfoVector{};
     }
@@ -273,18 +384,25 @@ StatusWith<MigrateInfoVector> BalancerChunkSelectionPolicyImpl::selectChunksToMo
         const NamespaceString nss(coll.getNs());
 
         if (!coll.getAllowBalance()) {
-            LOG(1) << "Not balancing collection " << nss << "; explicitly disabled.";
+            LOGV2_DEBUG(21851,
+                        1,
+                        "Not balancing collection {namespace}; explicitly disabled.",
+                        "Not balancing explicitly disabled collection",
+                        "namespace"_attr = nss);
             continue;
         }
 
-        auto candidatesStatus = _getMigrateCandidatesForCollection(
-            opCtx, nss, shardStats, aggressiveBalanceHint, &usedShards);
+        auto candidatesStatus =
+            _getMigrateCandidatesForCollection(opCtx, nss, shardStats, &usedShards);
         if (candidatesStatus == ErrorCodes::NamespaceNotFound) {
             // Namespace got dropped before we managed to get to it, so just skip it
             continue;
         } else if (!candidatesStatus.isOK()) {
-            warning() << "Unable to balance collection " << nss.ns()
-                      << causedBy(candidatesStatus.getStatus());
+            LOGV2_WARNING(21853,
+                          "Unable to balance collection {namespace}: {error}",
+                          "Unable to balance collection",
+                          "namespace"_attr = nss.ns(),
+                          "error"_attr = candidatesStatus.getStatus());
             continue;
         }
 
@@ -296,6 +414,38 @@ StatusWith<MigrateInfoVector> BalancerChunkSelectionPolicyImpl::selectChunksToMo
     return candidateChunks;
 }
 
+StatusWith<MigrateInfoVector> BalancerChunkSelectionPolicyImpl::selectChunksToMove(
+    OperationContext* opCtx, const NamespaceString& nss) {
+    auto shardStatsStatus = _clusterStats->getStats(opCtx);
+    if (!shardStatsStatus.isOK()) {
+        return shardStatsStatus.getStatus();
+    }
+
+    const auto& shardStats = shardStatsStatus.getValue();
+
+    // Validate collection information
+    auto swCollection = Grid::get(opCtx)->catalogClient()->getCollection(opCtx, nss);
+    if (!swCollection.isOK()) {
+        return swCollection.getStatus();
+    }
+
+    const auto& collection = swCollection.getValue().value;
+
+    if (collection.getDropped()) {
+        return Status(ErrorCodes::NamespaceNotFound,
+                      str::stream() << "collection " << nss.ns() << " not found");
+    }
+
+    std::set<ShardId> usedShards;
+
+    auto candidatesStatus = _getMigrateCandidatesForCollection(opCtx, nss, shardStats, &usedShards);
+    if (!candidatesStatus.isOK()) {
+        return candidatesStatus.getStatus();
+    }
+
+    return candidatesStatus;
+}
+
 StatusWith<boost::optional<MigrateInfo>>
 BalancerChunkSelectionPolicyImpl::selectSpecificChunkToMove(OperationContext* opCtx,
                                                             const ChunkType& chunk) {
@@ -307,8 +457,8 @@ BalancerChunkSelectionPolicyImpl::selectSpecificChunkToMove(OperationContext* op
     const auto& shardStats = shardStatsStatus.getValue();
 
     auto routingInfoStatus =
-        Grid::get(opCtx)->catalogCache()->getShardedCollectionRoutingInfoWithRefresh(
-            opCtx, NamespaceString(chunk.getNS()));
+        Grid::get(opCtx)->catalogCache()->getShardedCollectionRoutingInfoWithRefresh(opCtx,
+                                                                                     chunk.getNS());
     if (!routingInfoStatus.isOK()) {
         return routingInfoStatus.getStatus();
     }
@@ -336,8 +486,8 @@ Status BalancerChunkSelectionPolicyImpl::checkMoveAllowed(OperationContext* opCt
     auto shardStats = std::move(shardStatsStatus.getValue());
 
     auto routingInfoStatus =
-        Grid::get(opCtx)->catalogCache()->getShardedCollectionRoutingInfoWithRefresh(
-            opCtx, NamespaceString(chunk.getNS()));
+        Grid::get(opCtx)->catalogCache()->getShardedCollectionRoutingInfoWithRefresh(opCtx,
+                                                                                     chunk.getNS());
     if (!routingInfoStatus.isOK()) {
         return routingInfoStatus.getStatus();
     }
@@ -377,8 +527,6 @@ StatusWith<SplitInfoVector> BalancerChunkSelectionPolicyImpl::_getSplitCandidate
 
     const auto cm = routingInfoStatus.getValue().cm().get();
 
-    const auto& shardKeyPattern = cm->getShardKeyPattern().getKeyPattern();
-
     const auto collInfoStatus = createCollectionDistributionStatus(opCtx, shardStats, cm);
     if (!collInfoStatus.isOK()) {
         return collInfoStatus.getStatus();
@@ -389,28 +537,16 @@ StatusWith<SplitInfoVector> BalancerChunkSelectionPolicyImpl::_getSplitCandidate
     // Accumulate split points for the same chunk together
     SplitCandidatesBuffer splitCandidates(nss, cm->getVersion());
 
-    for (const auto& tagRangeEntry : distribution.tagRanges()) {
-        const auto& tagRange = tagRangeEntry.second;
-
-        const auto chunkAtZoneMin = cm->findIntersectingChunkWithSimpleCollation(tagRange.min);
-        invariant(chunkAtZoneMin->getMax().woCompare(tagRange.min) > 0);
-
-        if (chunkAtZoneMin->getMin().woCompare(tagRange.min)) {
-            splitCandidates.addSplitPoint(chunkAtZoneMin, tagRange.min);
+    if (nss == NamespaceString::kLogicalSessionsNamespace) {
+        if (!distribution.tags().empty()) {
+            LOGV2_WARNING(4562401,
+                          "Ignoring zones for the sessions collection",
+                          "tags"_attr = distribution.tags());
         }
 
-        // The global max key can never fall in the middle of a chunk
-        if (!tagRange.max.woCompare(shardKeyPattern.globalMax()))
-            continue;
-
-        const auto chunkAtZoneMax = cm->findIntersectingChunkWithSimpleCollation(tagRange.max);
-
-        // We need to check that both the chunk's minKey does not match the zone's max and also that
-        // the max is not equal, which would only happen in the case of the zone ending in MaxKey.
-        if (chunkAtZoneMax->getMin().woCompare(tagRange.max) &&
-            chunkAtZoneMax->getMax().woCompare(tagRange.max)) {
-            splitCandidates.addSplitPoint(chunkAtZoneMax, tagRange.max);
-        }
+        getSplitCandidatesForSessionsCollection(opCtx, cm, &splitCandidates);
+    } else {
+        getSplitCandidatesToEnforceTagRanges(cm, distribution, &splitCandidates);
     }
 
     return splitCandidates.done();
@@ -420,7 +556,6 @@ StatusWith<MigrateInfoVector> BalancerChunkSelectionPolicyImpl::_getMigrateCandi
     OperationContext* opCtx,
     const NamespaceString& nss,
     const ShardStatisticsVector& shardStats,
-    bool aggressiveBalanceHint,
     std::set<ShardId>* usedShards) {
     auto routingInfoStatus =
         Grid::get(opCtx)->catalogCache()->getShardedCollectionRoutingInfoWithRefresh(opCtx, nss);
@@ -444,15 +579,13 @@ StatusWith<MigrateInfoVector> BalancerChunkSelectionPolicyImpl::_getMigrateCandi
 
         const auto chunkAtZoneMin = cm->findIntersectingChunkWithSimpleCollation(tagRange.min);
 
-        if (chunkAtZoneMin->getMin().woCompare(tagRange.min)) {
+        if (chunkAtZoneMin.getMin().woCompare(tagRange.min)) {
             return {ErrorCodes::IllegalOperation,
                     str::stream()
-                        << "Tag boundaries "
-                        << tagRange.toString()
+                        << "Tag boundaries " << tagRange.toString()
                         << " fall in the middle of an existing chunk "
-                        << ChunkRange(chunkAtZoneMin->getMin(), chunkAtZoneMin->getMax()).toString()
-                        << ". Balancing for collection "
-                        << nss.ns()
+                        << ChunkRange(chunkAtZoneMin.getMin(), chunkAtZoneMin.getMax()).toString()
+                        << ". Balancing for collection " << nss.ns()
                         << " will be postponed until the chunk is split appropriately."};
         }
 
@@ -464,21 +597,23 @@ StatusWith<MigrateInfoVector> BalancerChunkSelectionPolicyImpl::_getMigrateCandi
 
         // We need to check that both the chunk's minKey does not match the zone's max and also that
         // the max is not equal, which would only happen in the case of the zone ending in MaxKey.
-        if (chunkAtZoneMax->getMin().woCompare(tagRange.max) &&
-            chunkAtZoneMax->getMax().woCompare(tagRange.max)) {
+        if (chunkAtZoneMax.getMin().woCompare(tagRange.max) &&
+            chunkAtZoneMax.getMax().woCompare(tagRange.max)) {
             return {ErrorCodes::IllegalOperation,
                     str::stream()
-                        << "Tag boundaries "
-                        << tagRange.toString()
+                        << "Tag boundaries " << tagRange.toString()
                         << " fall in the middle of an existing chunk "
-                        << ChunkRange(chunkAtZoneMax->getMin(), chunkAtZoneMax->getMax()).toString()
-                        << ". Balancing for collection "
-                        << nss.ns()
+                        << ChunkRange(chunkAtZoneMax.getMin(), chunkAtZoneMax.getMax()).toString()
+                        << ". Balancing for collection " << nss.ns()
                         << " will be postponed until the chunk is split appropriately."};
         }
     }
 
-    return BalancerPolicy::balance(shardStats, distribution, aggressiveBalanceHint, usedShards);
+    return BalancerPolicy::balance(
+        shardStats,
+        distribution,
+        usedShards,
+        Grid::get(opCtx)->getBalancerConfiguration()->attemptToBalanceJumboChunks());
 }
 
 }  // namespace mongo
